"""Model inference demonstration tool."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import time
from collections import deque

import cv2
import numpy as np

from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.datasets import DatasetEntity, DatasetItemEntity
from otx.api.entities.image import Image
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.task_environment import TaskEnvironment
from otx.cli.manager import ConfigManager
from otx.cli.tools.utils.demo.images_capture import open_images_capture
from otx.cli.tools.utils.demo.visualization import draw_predictions, put_text_on_rect_bg
from otx.cli.utils.importing import get_impl_class
from otx.cli.utils.io import read_label_schema, read_model
from otx.cli.utils.parser import (
    add_hyper_parameters_sub_parser,
    get_parser_and_hprams_data,
)

ESC_BUTTON = 27


def get_args():
    """Parses command line arguments."""
    parser, hyper_parameters, params = get_parser_and_hprams_data()

    parser.add_argument(
        "-i",
        "--input",
        required=True,
        help="Source of input data: images folder, image, webcam and video.",
    )
    parser.add_argument(
        "--load-weights",
        required=True,
        help="Load model weights from previously saved checkpoint."
        "It could be a trained/optimized model (POT only) or exported model.",
    )
    parser.add_argument(
        "--fit-to-size",
        nargs=2,
        type=int,
        help="Width and Height space-separated values. "
        "Fits displayed images to window with specified Width and Height. "
        "This options applies to result visualisation only.",
    )
    parser.add_argument("--loop", action="store_true", help="Enable reading the input in a loop.")
    parser.add_argument("--delay", type=int, default=3, help="Frame visualization time in ms.")
    parser.add_argument(
        "--display-perf",
        action="store_true",
        help="This option enables writing performance metrics on displayed frame. "
        "These metrics take into account not only model inference time, but also "
        "frame reading, pre-processing and post-processing.",
    )

    add_hyper_parameters_sub_parser(parser, hyper_parameters, modes=("INFERENCE",))
    override_param = [f"params.{param[2:].split('=')[0]}" for param in params if param.startswith("--")]

    return parser.parse_args(), override_param


def get_predictions(task, frame):
    """Returns list of predictions made by task on frame and time spent on doing prediction."""

    empty_annotation = AnnotationSceneEntity(annotations=[], kind=AnnotationSceneKind.PREDICTION)

    item = DatasetItemEntity(
        media=Image(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)),
        annotation_scene=empty_annotation,
    )

    dataset = DatasetEntity(items=[item])

    start_time = time.perf_counter()
    predicted_validation_dataset = task.infer(
        dataset,
        InferenceParameters(is_evaluation=True),
    )
    elapsed_time = time.perf_counter() - start_time
    item = predicted_validation_dataset[0]
    return item.get_annotations(), elapsed_time


def main():
    """Main function that is used for model demonstration."""

    # Dynamically create an argument parser based on override parameters.
    args, override_param = get_args()

    config_manager = ConfigManager(args, mode="eval")
    # Auto-Configuration for model template
    config_manager.configure_template()

    # Update Hyper Parameter Configs
    hyper_parameters = config_manager.get_hyparams_config(override_param)

    # Get classes for Task, ConfigurableParameters and Dataset.
    template = config_manager.template
    if any(args.load_weights.endswith(x) for x in (".bin", ".xml", ".zip")):
        task_class = get_impl_class(template.entrypoints.openvino)
    elif args.load_weights.endswith(".pth"):
        task_class = get_impl_class(template.entrypoints.base)
    else:
        raise ValueError(f"Unsupported file: {args.load_weights}")

    environment = TaskEnvironment(
        model=None,
        hyper_parameters=hyper_parameters,
        label_schema=read_label_schema(args.load_weights),
        model_template=template,
    )

    environment.model = read_model(environment.get_model_configuration(), args.load_weights, None)

    task = task_class(task_environment=environment)

    capture = open_images_capture(args.input, args.loop)

    elapsed_times = deque(maxlen=10)
    frame_index = 0
    while True:
        frame = capture.read()
        if frame is None:
            break

        predictions, elapsed_time = get_predictions(task, frame)
        elapsed_times.append(elapsed_time)
        elapsed_time = np.mean(elapsed_times)

        frame = draw_predictions(template.task_type, predictions, frame, args.fit_to_size)
        if args.display_perf:
            put_text_on_rect_bg(
                frame,
                f"time: {elapsed_time:.4f} sec.",
                (0, frame.shape[0] - 30),
                color=(255, 255, 255),
            )

        if args.delay >= 0:
            cv2.imshow("frame", frame)
            if cv2.waitKey(args.delay) == ESC_BUTTON:
                break
        else:
            print(f"{frame_index=}, {elapsed_time=}, {len(predictions)=}")

    return dict(retcode=0, template=template.name)


if __name__ == "__main__":
    main()



"""NNCF Task of OTX Detection."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


from functools import partial
from typing import Optional

import otx.algorithms.detection.adapters.mmdet.nncf.patches  # noqa: F401  # pylint: disable=unused-import
from otx.algorithms.common.adapters.mmcv.utils import remove_from_config
from otx.algorithms.common.tasks.nncf_base import NNCFBaseTask
from otx.algorithms.detection.adapters.mmdet.nncf import build_nncf_detector
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.mpa.utils.logger import get_logger

from .inference import DetectionInferenceTask
from .train import DetectionTrainTask

logger = get_logger()


class DetectionNNCFTask(NNCFBaseTask, DetectionInferenceTask):  # pylint: disable=too-many-ancestors
    """DetectionNNCFTask."""

    def _initialize_post_hook(self, options=None):
        super()._initialize_post_hook(options)

        export = options.get("export", False)
        options["model_builder"] = partial(
            self.model_builder,
            nncf_model_builder=build_nncf_detector,
            return_compression_ctrl=False,
            is_export=export,
        )

        # do not configure regularization
        if "l2sp_weight" in self._recipe_cfg.model or "l2sp_weight" in self._model_cfg.model:
            remove_from_config(self._recipe_cfg.model, "l2sp_weight")
            remove_from_config(self._model_cfg.model, "l2sp_weight")

    def _optimize(
        self,
        dataset: DatasetEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        results = self._run_task(
            "DetectionTrainer",
            mode="train",
            dataset=dataset,
            parameters=optimization_parameters,
        )

        return results

    def _optimize_post_hook(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
    ):
        # get prediction on validation set
        val_dataset = dataset.get_subset(Subset.VALIDATION)
        val_preds, val_map = self._infer_detector(val_dataset, InferenceParameters(is_evaluation=True))

        preds_val_dataset = val_dataset.with_empty_annotations()
        self._add_predictions_to_dataset(val_preds, preds_val_dataset, 0.0)

        result_set = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=val_dataset,
            prediction_dataset=preds_val_dataset,
        )

        # adjust confidence threshold
        if self._hyperparams.postprocessing.result_based_confidence_threshold:
            best_confidence_threshold = None
            logger.info("Adjusting the confidence threshold")
            metric = MetricsHelper.compute_f_measure(result_set, vary_confidence_threshold=True)
            if metric.best_confidence_threshold:
                best_confidence_threshold = metric.best_confidence_threshold.value
            if best_confidence_threshold is None:
                raise ValueError("Cannot compute metrics: Invalid confidence threshold!")
            logger.info(f"Setting confidence threshold to {best_confidence_threshold} based on results")
            self.confidence_threshold = best_confidence_threshold
        else:
            metric = MetricsHelper.compute_f_measure(result_set, vary_confidence_threshold=False)

        performance = metric.get_performance()
        logger.info(f"Final model performance: {str(performance)}")
        performance.dashboard_metrics.extend(
            # pylint: disable-next=protected-access
            DetectionTrainTask._generate_training_metrics(self._learning_curves, val_map)
        )
        output_model.performance = performance

    def _save_model_post_hook(self, modelinfo):
        config = modelinfo["meta"]["config"]
        if hasattr(config.model, "bbox_head") and hasattr(config.model.bbox_head, "anchor_generator"):
            if getattr(
                config.model.bbox_head.anchor_generator,
                "reclustering_anchors",
                False,
            ):
                generator = config.model.bbox_head.anchor_generator
                modelinfo["anchors"] = {
                    "heights": generator.heights,
                    "widths": generator.widths,
                }

        modelinfo["confidence_threshold"] = self.confidence_threshold



# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from collections import OrderedDict
from copy import deepcopy
from pprint import pformat
from typing import Any, Callable, Dict, List, Optional, Type

import pytest

from otx.algorithms.anomaly.adapters.anomalib.logger import get_logger
from otx.api.configuration.helper import create as otx_sdk_configuration_helper_create
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import TaskType, parse_model_template
from otx.api.entities.subset import Subset
from otx.api.entities.train_parameters import TrainParameters
from tests.regression.anomaly.anomaly_common import (
    _create_anomaly_dataset_and_labels_schema,
    _get_dataset_params_from_dataset_definitions,
    get_anomaly_domain_test_action_classes,
    get_dummy_compressed_model,
)
from tests.test_suite.e2e_test_system import DataCollector, e2e_pytest_performance
from tests.test_suite.training_test_case import (
    OTXTestCaseInterface,
    generate_otx_integration_test_case_class,
)
from tests.test_suite.training_tests_actions import (
    OTXTestTrainingAction,
    create_environment_and_task,
)
from tests.test_suite.training_tests_common import (
    KEEP_CONFIG_FIELD_VALUE,
    REALLIFE_USECASE_CONSTANT,
    ROOT_PATH_KEY,
    make_path_be_abs,
    performance_to_score_name_value,
)
from tests.test_suite.training_tests_helper import (
    DefaultOTXTestCreationParametersInterface,
    OTXTestHelper,
    OTXTrainingTestInterface,
)

logger = get_logger(__name__)


@pytest.fixture
def otx_test_domain_fx():
    return "custom-anomaly-classification"


class AnomalyClassificationTrainingTestParameters(DefaultOTXTestCreationParametersInterface):
    def test_case_class(self) -> Type[OTXTestCaseInterface]:
        return generate_otx_integration_test_case_class(
            get_anomaly_domain_test_action_classes(AnomalyDetectionTestTrainingAction)
        )

    def test_bunches(self) -> List[Dict[str, Any]]:
        test_bunches = [
            dict(
                model_name=[
                    "otx_anomaly_classification_padim",
                    "otx_anomaly_classification_stfpm",
                ],
                dataset_name="mvtec_short",
                usecase="precommit",
            ),
            dict(
                model_name=[
                    "otx_anomaly_classification_padim",
                    "otx_anomaly_classification_stfpm",
                ],
                dataset_name="mvtec",
                patience=KEEP_CONFIG_FIELD_VALUE,
                batch_size=KEEP_CONFIG_FIELD_VALUE,
                usecase=REALLIFE_USECASE_CONSTANT,
            ),
        ]
        return deepcopy(test_bunches)

    def short_test_parameters_names_for_generating_id(self) -> OrderedDict:
        DEFAULT_SHORT_TEST_PARAMETERS_NAMES_FOR_GENERATING_ID = OrderedDict(
            [
                ("test_stage", "ACTION"),
                ("model_name", "model"),
                ("dataset_name", "dataset"),
                ("patience", "patience"),
                ("batch_size", "batch"),
                ("usecase", "usecase"),
            ]
        )
        return deepcopy(DEFAULT_SHORT_TEST_PARAMETERS_NAMES_FOR_GENERATING_ID)

    def test_parameters_defining_test_case_behavior(self) -> List[str]:
        DEFAULT_TEST_PARAMETERS_DEFINING_IMPL_BEHAVIOR = [
            "model_name",
            "dataset_name",
            "patience",
            "batch_size",
        ]
        return deepcopy(DEFAULT_TEST_PARAMETERS_DEFINING_IMPL_BEHAVIOR)

    def default_test_parameters(self) -> Dict[str, Any]:
        DEFAULT_TEST_PARAMETERS = {
            "patience": 1,
            "batch_size": 2,
        }
        return deepcopy(DEFAULT_TEST_PARAMETERS)


# TODO:pfinashx: This function copies with minor change OTXTestTrainingAction
#             from tests.test_suite.
#             Try to avoid copying of code.
class AnomalyDetectionTestTrainingAction(OTXTestTrainingAction):
    _name = "training"

    def __init__(self, dataset, labels_schema, template_path, patience, batch_size):
        self.dataset = dataset
        self.labels_schema = labels_schema
        self.template_path = template_path
        self.num_training_iters = patience
        self.batch_size = batch_size

    def _get_training_performance_as_score_name_value(self):
        training_performance = getattr(self.output_model, "performance", None)
        if training_performance is None:
            raise RuntimeError("Cannot get training performance")
        return performance_to_score_name_value(training_performance)

    def _run_otx_training(self, data_collector):
        logger.debug(f"self.template_path = {self.template_path}")

        print(f"train dataset: {len(self.dataset.get_subset(Subset.TRAINING))} items")
        print(f"validation dataset: " f"{len(self.dataset.get_subset(Subset.VALIDATION))} items")

        logger.debug("Load model template")
        self.model_template = parse_model_template(self.template_path)

        logger.debug("Set hyperparameters")
        params = otx_sdk_configuration_helper_create(self.model_template.hyper_parameters.data)
        if hasattr(params, "learning_parameters") and hasattr(params.learning_parameters, "early_stopping"):
            if self.num_training_iters != KEEP_CONFIG_FIELD_VALUE:
                params.learning_parameters.early_stopping.patience = int(self.num_training_iters)
                logger.debug(
                    f"Set params.learning_parameters.early_stopping.patience="
                    f"{params.learning_parameters.early_stopping.patience}"
                )
            else:
                logger.debug(
                    f"Keep params.learning_parameters.early_stopping.patience="
                    f"{params.learning_parameters.early_stopping.patience}"
                )
        if self.batch_size != KEEP_CONFIG_FIELD_VALUE:
            params.learning_parameters.train_batch_size = int(self.batch_size)
            logger.debug(
                f"Set params.learning_parameters.train_batch_size=" f"{params.learning_parameters.train_batch_size}"
            )
        else:
            logger.debug(
                f"Keep params.learning_parameters.train_batch_size=" f"{params.learning_parameters.train_batch_size}"
            )

        logger.debug("Setup environment")
        self.environment, self.task = create_environment_and_task(params, self.labels_schema, self.model_template)

        logger.debug("Train model")
        self.output_model = ModelEntity(
            self.dataset,
            self.environment.get_model_configuration(),
        )

        self.copy_hyperparams = deepcopy(self.task.task_environment.get_hyper_parameters())

        try:
            # fix seed so that result is repeatable
            self.task.train(self.dataset, self.output_model, TrainParameters, seed=42)
        except Exception as ex:
            raise RuntimeError("Training failed") from ex

        score_name, score_value = self._get_training_performance_as_score_name_value()
        logger.info(f"performance={self.output_model.performance}")
        data_collector.log_final_metric("metric_name", self.name + "/" + score_name)
        data_collector.log_final_metric("metric_value", score_value)

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._run_otx_training(data_collector)
        results = {
            "model_template": self.model_template,
            "task": self.task,
            "dataset": self.dataset,
            "environment": self.environment,
            "output_model": self.output_model,
        }
        return results


class TestOTXReallifeAnomalyClassification(OTXTrainingTestInterface):
    """
    The main class of running test in this file.
    """

    PERFORMANCE_RESULTS = None  # it is required for e2e system
    helper = OTXTestHelper(AnomalyClassificationTrainingTestParameters())

    @classmethod
    def get_list_of_tests(cls, usecase: Optional[str] = None):
        """
        This method should be a classmethod. It is called before fixture initialization, during
        tests discovering.
        """
        return cls.helper.get_list_of_tests(usecase)

    @pytest.fixture
    def params_factories_for_test_actions_fx(
        self, current_test_parameters_fx, dataset_definitions_fx, otx_current_reference_dir_fx, template_paths_fx
    ) -> Dict[str, Callable[[], Dict]]:
        logger.debug("params_factories_for_test_actions_fx: begin")

        test_parameters = deepcopy(current_test_parameters_fx)
        dataset_definitions = deepcopy(dataset_definitions_fx)
        template_paths = deepcopy(template_paths_fx)

        def _training_params_factory() -> Dict:
            if dataset_definitions is None:
                pytest.skip('The parameter "--dataset-definitions" is not set')
            model_name = test_parameters["model_name"]
            dataset_name = test_parameters["dataset_name"]
            patience = test_parameters["patience"]
            batch_size = test_parameters["batch_size"]
            dataset_params = _get_dataset_params_from_dataset_definitions(dataset_definitions, dataset_name)

            if model_name not in template_paths:
                raise ValueError(
                    f"Model {model_name} is absent in template_paths, "
                    f"template_paths.keys={list(template_paths.keys())}"
                )
            template_path = make_path_be_abs(template_paths[model_name], template_paths[ROOT_PATH_KEY])
            logger.debug("training params factory: Before creating dataset and labels_schema")
            dataset, labels_schema = _create_anomaly_dataset_and_labels_schema(
                dataset_params, dataset_name, TaskType.ANOMALY_CLASSIFICATION
            )
            logger.debug("training params factory: After creating dataset and labels_schema")
            return {
                "dataset": dataset,
                "labels_schema": labels_schema,
                "template_path": template_path,
                "patience": patience,
                "batch_size": batch_size,
            }

        def _nncf_graph_params_factory() -> Dict:
            if dataset_definitions is None:
                pytest.skip('The parameter "--dataset-definitions" is not set')

            model_name = test_parameters["model_name"]
            dataset_name = test_parameters["dataset_name"]

            dataset_params = _get_dataset_params_from_dataset_definitions(dataset_definitions, dataset_name)

            if model_name not in template_paths:
                raise ValueError(
                    f"Model {model_name} is absent in template_paths, "
                    f"template_paths.keys={list(template_paths.keys())}"
                )
            template_path = make_path_be_abs(template_paths[model_name], template_paths[ROOT_PATH_KEY])

            logger.debug("training params factory: Before creating dataset and labels_schema")
            dataset, labels_schema = _create_anomaly_dataset_and_labels_schema(
                dataset_params, dataset_name, TaskType.ANOMALY_CLASSIFICATION
            )
            logger.debug("training params factory: After creating dataset and labels_schema")

            return {
                "dataset": dataset,
                "labels_schema": labels_schema,
                "template_path": template_path,
                "reference_dir": otx_current_reference_dir_fx,
                "fn_get_compressed_model": get_dummy_compressed_model,
            }

        def _reference_params_factory() -> Dict:
            return {"reference_dir": otx_current_reference_dir_fx}

        params_factories_for_test_actions = {
            "training": _training_params_factory,
            "nncf_graph": _nncf_graph_params_factory,
            "nncf_validate_fq": _reference_params_factory,
            "pot_validate_fq": _reference_params_factory,
        }
        logger.debug("params_factories_for_test_actions_fx: end")
        return params_factories_for_test_actions

    @pytest.fixture
    def test_case_fx(self, current_test_parameters_fx, params_factories_for_test_actions_fx):
        """
        This fixture returns the test case class OTXIntegrationTestCase that should be used for the current test.
        Notx that the cache from the test helper allows to store the instance of the class
        between the tests.
        If the main parameters used for this test are the same as the main parameters used for the previous test,
        the instance of the test case class will be kept and re-used. It is helpful for tests that can
        re-use the result of operations (model training, model optimization, etc) made for the previous tests,
        if these operations are time-consuming.
        If the main parameters used for this test differs w.r.t. the previous test, a new instance of
        test case class will be created.
        """
        test_case = type(self).helper.get_test_case(current_test_parameters_fx, params_factories_for_test_actions_fx)
        return test_case

    # TODO(lbeynens): move to common fixtures
    @pytest.fixture
    def data_collector_fx(self, request) -> DataCollector:
        setup = deepcopy(request.node.callspec.params)
        setup["environment_name"] = os.environ.get("TT_ENVIRONMENT_NAME", "no-env")
        setup["test_type"] = os.environ.get("TT_TEST_TYPE", "no-test-type")  # TODO: get from e2e test type
        setup["scenario"] = "api"  # TODO(lbeynens): get from a fixture!
        setup["test"] = request.node.name
        setup["subject"] = "custom-anomaly-classification"
        setup["project"] = "otx"
        if "test_parameters" in setup:
            assert isinstance(setup["test_parameters"], dict)
            if "dataset_name" not in setup:
                setup["dataset_name"] = setup["test_parameters"].get("dataset_name")
            if "model_name" not in setup:
                setup["model_name"] = setup["test_parameters"].get("model_name")
            if "test_stage" not in setup:
                setup["test_stage"] = setup["test_parameters"].get("test_stage")
            if "usecase" not in setup:
                setup["usecase"] = setup["test_parameters"].get("usecase")
        logger.info(f"creating DataCollector: setup=\n{pformat(setup, width=140)}")
        data_collector = DataCollector(name="TestOTXIntegration", setup=setup)
        with data_collector:
            logger.info("data_collector is created")
            yield data_collector
        logger.info("data_collector is released")

    @e2e_pytest_performance
    def test(self, test_parameters, test_case_fx, data_collector_fx, cur_test_expected_metrics_callback_fx):
        test_case_fx.run_stage(test_parameters["test_stage"], data_collector_fx, cur_test_expected_metrics_callback_fx)



"""DetCon loss."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=no-name-in-module, not-callable
import itertools

import torch
import torch.distributed as dist
import torch.nn.functional as F
from mmseg.models.builder import LOSSES
from torch import nn


def manual_cross_entropy(logits, labels, weight):
    """Manually calculate weighted cross entropy."""
    cross_entropy = -weight * torch.sum(labels * F.log_softmax(logits, dim=-1), dim=-1)
    return torch.mean(cross_entropy)


@LOSSES.register_module
class DetConLoss(nn.Module):
    """Modified from https://github.com/deepmind/detcon/blob/main/utils/losses.py.

    Compute the NCE scores from pairs of predictions and targets.
    This implements the batched form of the loss described in
    Section 3.1, Equation 3 in https://arxiv.org/pdf/2103.10957.pdf.

    Args:
        temperature: (float) the temperature to use for the NCE loss.
        use_replicator_loss (bool): use cross-replica samples.
    """

    def __init__(
        self, temperature: float = 0.1, use_replicator_loss: bool = True, ignore_index: int = 255
    ):  # pylint: disable=unused-argument
        super().__init__()
        assert temperature > 0
        self.temperature = torch.tensor(temperature)
        self.use_replicator_loss = use_replicator_loss

    def get_distributed_tensors(self, target1, target2, batch_size, num_samples, num_features, device):
        """Grab tensors across replicas during distributed training."""
        num_gpus = torch.cuda.device_count()
        if num_gpus > 1 and torch.distributed.is_initialized() and self.use_replicator_loss:
            # Grab tensor across replicas and expand first dimension
            target1_large = [torch.zeros_like(target1) for _ in range(num_gpus)]
            target2_large = [torch.zeros_like(target2) for _ in range(num_gpus)]
            dist.all_gather(target1_large, target1)
            dist.all_gather(target2_large, target2)
            target1_large = torch.cat(target1_large, dim=0)
            target2_large = torch.cat(target2_large, dim=0)

            # Fold into batch dimension
            target1_large = target1_large.reshape(-1, num_samples, num_features)
            target2_large = target2_large.reshape(-1, num_samples, num_features)

            # Create the labels by using the current replica ID and offsetting.
            replica_id = dist.get_rank()
            labels_idx = torch.arange(batch_size) + replica_id * batch_size
            enlarged_bs = target1_large.shape[0]
            labels = F.one_hot(labels_idx, num_classes=enlarged_bs).to(device)
        else:
            target1_large = target1
            target2_large = target2
            labels = F.one_hot(torch.arange(batch_size), num_classes=batch_size).to(device)

        labels = labels.unsqueeze(dim=2).unsqueeze(dim=1)

        return target1_large, target2_large, labels

    # pylint: disable=too-many-arguments, too-many-locals
    def forward(self, pred1, pred2, target1, target2, pind1, pind2, tind1, tind2, local_negatives=True):
        """Forward loss.

        Args:
            pred1 (Tensor): (b, num_samples, d) the prediction from first view.
            pred2 (Tensor): (b, num_samples, d) the prediction from second view.
            target1 (Tensor): (b, num_samples, d) the projection from first view.
            target2 (Tensor): (b, num_samples, d) the projection from second view.
            pind1 (Tensor): (b, num_samples) mask indices for first view's prediction.
            pind2 (Tensor): (b, num_samples) mask indices for second view's prediction.
            tind1 (Tensor): (b, num_samples) mask indices for first view's projection.
            tind2 (Tensor): (b, num_samples) mask indices for second view's projection.
            local_negatives (bool): whether to include local negatives.

        Returns:
            dict[str, Tensor]: A single scalar loss for the XT-NCE objective.
        """
        batch_size, num_samples, num_features = pred1.shape
        main_dtype = pred1.dtype
        # infinity_proxy is reduced to avoid overflow when training w/ fp16.
        infinity_proxy = 1e4  # Used for masks to proxy a very large number.

        def make_same_obj(ind_0, ind_1):
            same_obj = torch.eq(
                ind_0.reshape([batch_size, num_samples, 1]), ind_1.reshape([batch_size, 1, num_samples])
            )
            same_obj = same_obj.unsqueeze(2).to(main_dtype)
            return same_obj

        same_obj_dict = {}
        for pair, (pind, tind) in zip(
            ["aa", "ab", "ba", "bb"], list(itertools.product([pind1, pind2], [tind1, tind2]))
        ):
            same_obj_dict[pair] = make_same_obj(pind, tind)

        # L2 normalize the tensors to use for the cosine-similarity
        def normalize_same_dtype(logit, p=2, dim=1, eps=1e-12, dtype=None):
            # modified from torch.nn.functional.normalize
            denom = logit.norm(p, dim, keepdim=True, dtype=dtype).clamp_min(eps).expand_as(logit)
            return logit / denom

        pred1 = normalize_same_dtype(pred1, dim=-1, dtype=main_dtype)
        pred2 = normalize_same_dtype(pred2, dim=-1, dtype=main_dtype)
        target1 = normalize_same_dtype(target1, dim=-1, dtype=main_dtype)
        target2 = normalize_same_dtype(target2, dim=-1, dtype=main_dtype)
        target1_large, target2_large, labels = self.get_distributed_tensors(
            target1, target2, batch_size, num_samples, num_features, pred1.device
        )

        # Do our matmuls and mask out appropriately.
        logits_dict = {}
        for pair, (pred, target) in zip(
            ["aa", "ab", "ba", "bb"], list(itertools.product([pred1, pred2], [target1_large, target2_large]))
        ):
            logits_dict[pair] = torch.einsum("abk,uvk->abuv", pred, target) / self.temperature

        labels_dict = {key: labels * same_obj for key, same_obj in same_obj_dict.items()}
        for pair in ["aa", "bb"]:
            logits_dict[pair] -= infinity_proxy * labels * same_obj_dict[pair]
            labels_dict[pair] *= 0.0

        if not local_negatives:
            for pair in ["aa", "ab", "ba", "bb"]:
                logits_dict[pair] -= infinity_proxy * labels * (1 - same_obj_dict[pair])

        labels_concat = [
            torch.cat([labels_dict["ab"], labels_dict["aa"]], dim=2).reshape((batch_size, num_samples, -1)),
            torch.cat([labels_dict["ba"], labels_dict["bb"]], dim=2).reshape((batch_size, num_samples, -1)),
        ]

        num_positives = [torch.sum(label_concat, dim=-1, keepdim=True) for label_concat in labels_concat]

        labels_concat = [
            label_concat / torch.maximum(num_positive, torch.tensor(1.0, device=num_positive.device))
            for label_concat, num_positive in zip(labels_concat, num_positives)
        ]

        obj_areas = [torch.sum(make_same_obj(pind, pind), dim=(2, 3)) for pind in [pind1, pind2]]

        weights = [
            torch.greater(num_positive[..., 0], 1e-3).to(torch.float32) / obj_area
            for num_positive, obj_area in zip(num_positives, obj_areas)
        ]

        logits_concat = [
            torch.cat([logits_dict["ab"], logits_dict["aa"]], dim=2).reshape((batch_size, num_samples, -1)),
            torch.cat([logits_dict["ba"], logits_dict["bb"]], dim=2).reshape((batch_size, num_samples, -1)),
        ]

        loss_a = manual_cross_entropy(logits_concat[0], labels_concat[0], weight=weights[0])
        loss_b = manual_cross_entropy(logits_concat[1], labels_concat[1], weight=weights[1])
        loss = loss_a + loss_b

        return dict(loss=loss)



"""OTX MVTec Dataset facilitate OTX Anomaly Training.

License:
    MVTec AD dataset is released under the Creative Commons
    Attribution-NonCommercial-ShareAlike 4.0 International License
    (CC BY-NC-SA 4.0)(https://creativecommons.org/licenses/by-nc-sa/4.0/).

Reference:
    - Paul Bergmann, Kilian Batzner, Michael Fauser, David Sattlegger, Carsten Steger:
      The MVTec Anomaly Detection Dataset: A Comprehensive Real-World Dataset for
      Unsupervised Anomaly Detection; in: International Journal of Computer Vision
      129(4):1038-1059, 2021, DOI: 10.1007/s11263-020-01400-4.

    - Paul Bergmann, Michael Fauser, David Sattlegger, Carsten Steger: MVTec AD —
      A Comprehensive Real-World Dataset for Unsupervised Anomaly Detection;
      in: IEEE/CVF Conference on Computer Vision and Pattern Recognition (CVPR),
      9584-9592, 2019, DOI: 10.1109/CVPR.2019.00982.
"""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pathlib import Path
from typing import List, Union

import cv2
import numpy as np
from anomalib.data.mvtec import make_mvtec_dataset
from pandas.core.frame import DataFrame

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.color import Color
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.utils.segmentation_utils import create_annotation_from_segmentation_map


class OtxMvtecDataset:
    """Generate OTX MVTec Dataset from the anomaly detection datasets that follows the MVTec format.

    Args:
        path (Union[str, Path], optional): Path to the MVTec dataset category.
            Defaults to "./datasets/MVTec/bottle".
        split_ratio (float, optional): Ratio to split normal training images and add to the
            test set in case test set doesn't contain any normal images.
            Defaults to 0.5.
        seed (int, optional): Random seed to ensure reproducibility when splitting. Defaults to 0.
        create_validation_set (bool, optional): Create validation set from the test set by splitting
            it to half. Default to True.

    Examples:
        >>> dataset_generator = OtxMvtecDataset()
        >>> dataset = dataset_generator.generate()
        >>> dataset[0].media.numpy.shape
        (900, 900, 3)
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        path: Union[str, Path],
        task_type: TaskType = TaskType.ANOMALY_CLASSIFICATION,
    ):
        self.path = path if isinstance(path, Path) else Path(path)
        self.task_type = task_type

        if self.task_type == TaskType.ANOMALY_CLASSIFICATION:
            self.label_domain = Domain.ANOMALY_CLASSIFICATION
        elif self.task_type == TaskType.ANOMALY_SEGMENTATION:
            self.label_domain = Domain.ANOMALY_SEGMENTATION

        self.normal_label = LabelEntity(name="Normal", domain=self.label_domain, id=ID(), color=Color(0, 255, 0))
        self.abnormal_label = LabelEntity(
            name="Anomalous",
            domain=self.label_domain,
            id=ID(),
            is_anomalous=True,
            color=Color(255, 0, 0),
        )
        self.label_map = {0: self.normal_label, 1: self.abnormal_label}

    def get_samples(self) -> DataFrame:
        """Get MVTec samples.

        Get MVTec samples in a pandas DataFrame. Update the certain columns
        to match the OTX naming terminology. For example, column `split` is
        renamed to `subset`. Labels are also renamed by creating their
        corresponding OTX LabelEntities

        Returns:
            DataFrame: Final list of samples comprising all the required
                information to create the OTX Dataset.
        """
        samples = make_mvtec_dataset(root=self.path)

        # Set the OTX SDK Splits
        samples = samples.rename(columns={"split": "subset"})
        samples.loc[samples.subset == "train", "subset"] = Subset.TRAINING
        samples.loc[samples.subset == "val", "subset"] = Subset.VALIDATION
        samples.loc[samples.subset == "test", "subset"] = Subset.TESTING

        # Create and Set the OTX Labels
        samples.loc[samples.label != "good", "label"] = self.abnormal_label
        samples.loc[samples.label == "good", "label"] = self.normal_label

        samples = samples.reset_index(drop=True)

        return samples

    def generate(self) -> DatasetEntity:
        """Generate OTX Anomaly Dataset.

        Returns:
            DatasetEntity: Output OTX Anomaly Dataset from an MVTec
        """
        samples = self.get_samples()
        dataset_items: List[DatasetItemEntity] = []
        for _, sample in samples.iterrows():
            # Create image
            image = Image(file_path=sample.image_path)

            # Create annotation
            if self.task_type == TaskType.ANOMALY_CLASSIFICATION or sample.label == self.normal_label:
                shape = Rectangle(x1=0, y1=0, x2=1, y2=1)
                labels = [ScoredLabel(sample.label)]
                annotations = [Annotation(shape=shape, labels=labels)]
                annotation_scene = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)
            elif self.task_type == TaskType.ANOMALY_SEGMENTATION and sample.label == self.abnormal_label:
                mask = (cv2.imread(sample.mask_path, cv2.IMREAD_GRAYSCALE) / 255).astype(np.uint8)
                annotations = create_annotation_from_segmentation_map(mask, np.ones_like(mask), self.label_map)
                annotation_scene = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)
            else:
                raise ValueError(f"Unknown task type: {self.task_type}")

            # Create dataset item
            dataset_item = DatasetItemEntity(media=image, annotation_scene=annotation_scene, subset=sample.subset)

            # Add to dataset items
            dataset_items.append(dataset_item)

        dataset = DatasetEntity(items=dataset_items)
        return dataset



"""NMS Module."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np


def nms(boxes, scores, thresh):
    """Adapted NMS implementation from OMZ: model_zoo/model_api/models/utils.py#L181."""
    # pylint: disable=too-many-locals

    x1, y1, x2, y2 = boxes.T
    areas = (x2 - x1) * (y2 - y1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)

        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        width = np.maximum(0.0, xx2 - xx1)
        height = np.maximum(0.0, yy2 - yy1)
        intersection = width * height

        union = areas[i] + areas[order[1:]] - intersection
        overlap = np.divide(
            intersection,
            union,
            out=np.zeros_like(intersection, dtype=float),
            where=union != 0,
        )

        order = order[np.where(overlap <= thresh)[0] + 1]

    return keep


def multiclass_nms(
    detections: np.ndarray,
    iou_threshold=0.45,
    max_num=200,
):
    """Multi-class NMS.

    strategy: in order to perform NMS independently per class,
    we add an offset to all the boxes. The offset is dependent
    only on the class idx, and is large enough so that boxes
    from different classes do not overlap

    Args:
        detections (np.ndarray): labels, scores and boxes
        iou_threshold (float, optional): IoU threshold. Defaults to 0.45.
        max_num (int, optional): Max number of objects filter. Defaults to 200.

    Returns:
        _type_: _description_
    """
    labels = detections[:, 0]
    scores = detections[:, 1]
    boxes = detections[:, 2:]
    max_coordinate = boxes.max()
    offsets = labels.astype(boxes.dtype) * (max_coordinate + 1)
    boxes_for_nms = boxes + offsets[:, None]
    keep = nms(boxes_for_nms, scores, iou_threshold)
    if max_num > 0:
        keep = keep[:max_num]
    keep = np.array(keep)
    det = detections[keep]
    return det, keep



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.detection.adapters.mmdet.data.pipelines import (
    LoadAnnotationFromOTXDataset,
    LoadImageFromOTXDataset,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


class TestLoadImageFromOTXDatasetInputParamsValidation:
    @e2e_pytest_unit
    def test_load_image_from_otx_dataset_init_params_validation(self):
        """
        <b>Description:</b>
        Check LoadImageFromOTXDataset object initialization parameters validation

        <b>Input data:</b>
        "to_float32" non-bool parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        LoadImageFromOTXDataset object initialization parameter
        """
        with pytest.raises(ValueError):
            LoadImageFromOTXDataset("unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_load_image_from_otx_dataset_call_params_validation(self):
        """
        <b>Description:</b>
        Check LoadImageFromOTXDataset object "__call__" method input parameters validation

        <b>Input data:</b>
        LoadImageFromOTXDataset object, "results" unexpected type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "__call__" method
        """
        load_image_from_otx_dataset = LoadImageFromOTXDataset()
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "results" parameter
            unexpected_int,
            # Unexpected integer is specified as "results" dictionary key
            {"result_1": "some results", unexpected_int: "unexpected results"},
        ]:
            with pytest.raises(ValueError):
                load_image_from_otx_dataset.__call__(results=unexpected_value)


class TestLoadAnnotationFromOTXDatasetInputParamsValidation:
    @e2e_pytest_unit
    def test_load_annotation_from_otx_dataset_init_params_validation(self):
        """
        <b>Description:</b>
        Check LoadAnnotationFromOTXDataset object initialization parameters validation

        <b>Input data:</b>
        LoadAnnotationFromOTXDataset object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        LoadAnnotationFromOTXDataset object initialization parameter
        """
        correct_values_dict = {
            "min_size": 1,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "min_size" parameter
            ("min_size", unexpected_str),
            # Unexpected string is specified as "with_bbox" parameter
            ("with_bbox", unexpected_str),
            # Unexpected string is specified as "with_label" parameter
            ("with_label", unexpected_str),
            # Unexpected string is specified as "with_mask" parameter
            ("with_mask", unexpected_str),
            # Unexpected string is specified as "with_seg" parameter
            ("with_seg", unexpected_str),
            # Unexpected string is specified as "poly2mask" parameter
            ("poly2mask", unexpected_str),
            # Unexpected string is specified as "with_text" parameter
            ("with_text", unexpected_str),
            # Unexpected string is specified as "domain" parameter
            ("domain", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=LoadAnnotationFromOTXDataset,
        )

    @e2e_pytest_unit
    def test_load_annotation_from_otx_dataset_call_params_validation(self):
        """
        <b>Description:</b>
        Check LoadAnnotationFromOTXDataset object "__call__" method input parameters validation

        <b>Input data:</b>
        LoadAnnotationFromOTXDataset object, "results" unexpected type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "__call__" method
        """
        load_annotation_from_otx_dataset = LoadAnnotationFromOTXDataset(min_size=1)
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "results" parameter
            unexpected_int,
            # Unexpected integer is specified as "results" dictionary key
            {"result_1": "some results", unexpected_int: "unexpected results"},
        ]:
            with pytest.raises(ValueError):
                load_annotation_from_otx_dataset(results=unexpected_value)



__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
__img_scale = (544, 544)
__crop_size = (512, 512)

train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations"),
    dict(type="Resize", img_scale=__img_scale, ratio_range=(0.5, 2.0), keep_ratio=False),
    dict(type="RandomCrop", crop_size=__crop_size, cat_max_ratio=0.75),
    dict(type="RandomFlip", prob=0.5, direction="horizontal"),
    dict(
        type="MaskCompose",
        prob=0.5,
        lambda_limits=(4, 16),
        keep_original=False,
        transforms=[
            dict(type="PhotoMetricDistortion"),
        ],
    ),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="Pad", size=__crop_size, pad_val=0, seg_pad_val=255),
    dict(type="RandomRotate", prob=0.5, degree=30, pad_val=0, seg_pad_val=255),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_semantic_seg"]),
]
test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=__img_scale,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="RandomFlip"),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import PIL.Image
import pytest

from otx.algorithms.segmentation.adapters.mmseg.data.pipelines import (
    LoadAnnotationFromOTXDataset,
    NDArrayToPILImage,
    PILImageToNDArray,
    RandomResizedCrop,
    RandomSolarization,
)
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def label_entity(name="test label") -> LabelEntity:
    return LabelEntity(name=name, domain=Domain.SEGMENTATION)


def dataset_item() -> DatasetItemEntity:
    image: Image = Image(data=np.random.randint(low=0, high=255, size=(10, 16, 3)))
    annotation: Annotation = Annotation(shape=Rectangle.generate_full_box(), labels=[ScoredLabel(label_entity())])
    annotation_scene: AnnotationSceneEntity = AnnotationSceneEntity(
        annotations=[annotation], kind=AnnotationSceneKind.ANNOTATION
    )
    return DatasetItemEntity(media=image, annotation_scene=annotation_scene)


class TestLoadAnnotationFromOTXDataset:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:

        self.dataset_item: DatasetItemEntity = dataset_item()
        self.results: dict = {
            "dataset_item": self.dataset_item,
            "ann_info": {"labels": [label_entity("class_1")]},
            "seg_fields": [],
        }
        self.pipeline: LoadAnnotationFromOTXDataset = LoadAnnotationFromOTXDataset()

    @e2e_pytest_unit
    def test_call(self) -> None:
        loaded_annotations: dict = self.pipeline(self.results)
        assert "gt_semantic_seg" in loaded_annotations
        assert loaded_annotations["dataset_item"] == self.dataset_item


class TestNDArrayToPILImage:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.results: dict = {"img": np.random.randint(0, 255, (3, 3, 3), dtype=np.uint8)}
        self.nd_array_to_pil_image: NDArrayToPILImage = NDArrayToPILImage(keys=["img"])

    @e2e_pytest_unit
    def test_call(self) -> None:
        converted_img: dict = self.nd_array_to_pil_image(self.results)
        assert "img" in converted_img
        assert isinstance(converted_img["img"], PIL.Image.Image)

    @e2e_pytest_unit
    def test_repr(self) -> None:
        assert str(self.nd_array_to_pil_image) == "NDArrayToPILImage"


class TestPILImageToNDArray:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.results: dict = {"img": PIL.Image.new("RGB", (3, 3))}
        self.pil_image_to_nd_array: PILImageToNDArray = PILImageToNDArray(keys=["img"])

    @e2e_pytest_unit
    def test_call(self) -> None:
        converted_array: dict = self.pil_image_to_nd_array(self.results)
        assert "img" in converted_array
        assert isinstance(converted_array["img"], np.ndarray)

    @e2e_pytest_unit
    def test_repr(self) -> None:
        assert str(self.pil_image_to_nd_array) == "PILImageToNDArray"


class TestRandomResizedCrop:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.results: dict = {"img": PIL.Image.new("RGB", (10, 16)), "img_shape": (10, 16), "ori_shape": (10, 16)}
        self.random_resized_crop: RandomResizedCrop = RandomResizedCrop((5, 5), (0.5, 1.0))

    @e2e_pytest_unit
    def test_call(self) -> None:
        cropped_img: dict = self.random_resized_crop(self.results)
        assert cropped_img["img_shape"] == (5, 5)
        assert cropped_img["ori_shape"] == (10, 16)


class TestRandomSolarization:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.results: dict = {"img": np.random.randint(0, 255, (3, 3, 3), dtype=np.uint8)}
        self.random_solarization: RandomSolarization = RandomSolarization(p=1.0)

    @e2e_pytest_unit
    def test_call(self) -> None:
        solarized: dict = self.random_solarization(self.results)
        assert "img" in solarized
        assert isinstance(solarized["img"], np.ndarray)

    @e2e_pytest_unit
    def test_repr(self) -> None:
        assert str(self.random_solarization) == "RandomSolarization"



"""OTX Adapters for action recognition detectors - mmaction2."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from .fast_rcnn import AVAFastRCNN

__all__ = ["AVAFastRCNN"]



"""Dummy lightning modules for testing."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from typing import Any

import pytorch_lightning as pl
import torch
from anomalib.data.utils import masks_to_boxes
from anomalib.utils.metrics import AnomalyScoreThreshold, MinMax


class DummyModel(pl.LightningModule):
    """Returns mock outputs."""

    def __init__(self):
        super().__init__()
        self.image_threshold = AnomalyScoreThreshold()
        self.normalization_metrics = MinMax()

    def configure_optimizers(self):
        return None

    def training_step(self, *args, **kwargs):
        pass

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
        # Just return everything as anomalous
        batch["anomaly_maps"] = batch["pred_masks"] = torch.ones(batch["image"].shape[0], 1, *batch["image"].shape[2:])
        batch["pred_labels"] = batch["pred_scores"] = torch.ones((batch["image"].shape[0]))
        batch["pred_boxes"], batch["box_scores"] = masks_to_boxes(batch["pred_masks"], batch["anomaly_maps"])
        is_anomalous = [scores > 0.5 for scores in batch["box_scores"]]
        batch["box_labels"] = [labels.int() for labels in is_anomalous]
        return batch

    def validation_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0):
        return self.predict_step(batch, batch_idx, dataloader_idx)

    def test_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0):
        return self.predict_step(batch, batch_idx, dataloader_idx)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from copy import deepcopy
from functools import partial, reduce
from typing import List, Union

import pytest
import torch
import torch.nn as nn

from otx.algorithms.segmentation.adapters.mmseg import DetConB, SupConDetConB
from otx.algorithms.segmentation.adapters.mmseg.models.segmentors.detcon import (
    MaskPooling,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@pytest.fixture(autouse=True)
def setup_module(monkeypatch, mocker):
    class MockBackbone(nn.Module):
        def __init__(self):
            super().__init__()
            self.pipeline1 = nn.Sequential(nn.Conv2d(3, 1, (1, 1), bias=False), nn.Conv2d(1, 1, (1, 1), bias=False))
            self.pipeline2 = nn.Sequential(
                nn.Conv2d(3, 1, (1, 1), stride=2, bias=False), nn.Conv2d(1, 1, (1, 1), bias=False)
            )

        def init_weights(self, init_linear=None):
            pass

        def forward(self, x):
            return [self.pipeline1(x), self.pipeline2(x)]

    class MockNeck(nn.Sequential):
        def __init__(self):
            super().__init__(nn.Linear(2, 2, bias=False), nn.Linear(2, 2, bias=False))

        def init_weights(self, init_linear=None):
            pass

    class MockDecodeHead(nn.Module):
        def __init__(self):
            super().__init__()
            self.align_corners = None
            self.num_classes = 1
            self.out_channels = 1

        def forward(self, x):
            return x

    def build_mock(mock_class, *args, **kwargs):
        return mock_class()

    monkeypatch.setattr(
        "otx.algorithms.segmentation.adapters.mmseg.models.segmentors.detcon.build_backbone",
        partial(build_mock, MockBackbone),
    )
    monkeypatch.setattr(
        "otx.algorithms.segmentation.adapters.mmseg.models.segmentors.detcon.build_neck", partial(build_mock, MockNeck)
    )
    monkeypatch.setattr(
        "mmseg.models.segmentors.encoder_decoder.builder.build_backbone", partial(build_mock, MockBackbone)
    )
    monkeypatch.setattr(
        "mmseg.models.segmentors.encoder_decoder.builder.build_head", partial(build_mock, MockDecodeHead)
    )
    mocker.patch(
        "otx.algorithms.segmentation.adapters.mmseg.models.segmentors.detcon.build_loss",
        return_value=lambda *args, **kwargs: dict(loss=1.0),
    )
    mocker.patch(
        "otx.algorithms.segmentation.adapters.mmseg.models.segmentors.detcon.DetConB._register_state_dict_hook"
    )
    mocker.patch("otx.algorithms.segmentation.adapters.mmseg.models.segmentors.detcon.DetConB.state_dict_hook")
    mocker.patch("otx.algorithms.segmentation.adapters.mmseg.models.segmentors.detcon.load_checkpoint")
    mocker.patch(
        "otx.algorithms.segmentation.adapters.mmseg.models.segmentors.detcon.SupConDetConB._decode_head_forward_train",
        return_value=(dict(loss=1.0), None),
    )


class TestMaskPooling:
    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "masks", [torch.Tensor([[[[0, 1], [1, 2]]]]).to(torch.int64), torch.Tensor([[[0, 1], [1, 2]]]).to(torch.int64)]
    )
    @pytest.mark.parametrize("num_classes", [256, 3])
    def test_pool_masks(self, masks, num_classes):
        """Test pool_masks function."""
        mask_pooling = MaskPooling(num_classes=num_classes, downsample=1)

        binary_masks = mask_pooling.pool_masks(masks)

        assert binary_masks.shape[1] == num_classes

    @e2e_pytest_unit
    def test_sample_masks(self):
        """Test sample_masks function."""
        num_classes = 3
        masks = torch.Tensor([[[[0, 1], [1, 2]]]]).to(torch.int64)
        mask_pooling = MaskPooling(num_classes=num_classes, downsample=1)

        binary_masks = mask_pooling.pool_masks(masks)
        _, sampled_mask_ids = mask_pooling.sample_masks(binary_masks)

        assert len(torch.unique(sampled_mask_ids)) > 1
        assert len(torch.unique(sampled_mask_ids)) <= num_classes

    @e2e_pytest_unit
    def test_forward(self):
        """Test forward function."""
        num_classes = 3
        masks = torch.Tensor([[[[0, 1], [1, 2]]]]).to(torch.int64)
        mask_pooling = MaskPooling(num_classes=num_classes, downsample=1)

        sampled_masks, _ = mask_pooling(masks)

        assert torch.all(sampled_masks.sum(dim=-1) == 1)


class TestDetConB:
    """Test DetConB."""

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.detconb = DetConB(backbone={}, neck={}, head={}, downsample=1, loss_cfg=dict(type="DetConLoss"))

    @e2e_pytest_unit
    def test_init_weights(self) -> None:
        """Test init_weights function."""
        for param_ol, param_tgt in zip(
            self.detconb.online_backbone.parameters(), self.detconb.target_backbone.parameters()
        ):
            assert torch.all(param_ol == param_tgt)
            assert param_ol.requires_grad
            assert not param_tgt.requires_grad

        for param_ol, param_tgt in zip(
            self.detconb.online_projector.parameters(), self.detconb.target_projector.parameters()
        ):
            assert torch.all(param_ol == param_tgt)
            assert param_ol.requires_grad
            assert not param_tgt.requires_grad

    @e2e_pytest_unit
    def test_momentum_update(self) -> None:
        """Test _momentum_update function."""
        original_params = {"backbone": [], "projector": []}
        for param_tgt in self.detconb.target_backbone.parameters():
            param_tgt.data *= 2.0
            original_params["backbone"].append(deepcopy(param_tgt))

        for param_tgt in self.detconb.target_projector.parameters():
            param_tgt.data *= 2.0
            original_params["projector"].append(deepcopy(param_tgt))

        self.detconb._momentum_update()

        for param_ol, param_tgt, orig_tgt in zip(
            self.detconb.online_backbone.parameters(),
            self.detconb.target_backbone.parameters(),
            original_params["backbone"],
        ):
            assert torch.all(
                param_tgt.data == orig_tgt * self.detconb.momentum + param_ol.data * (1.0 - self.detconb.momentum)
            )

        for param_ol, param_tgt, orig_tgt in zip(
            self.detconb.online_projector.parameters(),
            self.detconb.target_projector.parameters(),
            original_params["projector"],
        ):
            assert torch.all(
                param_tgt.data == orig_tgt * self.detconb.momentum + param_ol.data * (1.0 - self.detconb.momentum)
            )

    @e2e_pytest_unit
    @pytest.mark.parametrize("input_transform", ["resize_concat", "multiple_select", "etc"])
    @pytest.mark.parametrize(
        "inputs,in_index",
        [
            ([torch.ones(1, 1, 4, 4), torch.ones(1, 1, 2, 2)], [0, 1]),
            ([torch.ones(1, 1, 4, 4), torch.ones(1, 1, 2, 2)], [0]),
            ([torch.ones(1, 1, 4, 4)], [0]),
        ],
    )
    def test_transform_inputs(self, inputs: List, in_index: Union[List, int], input_transform: str) -> None:
        """Test transform_inputs function."""
        setattr(self.detconb, "in_index", in_index)
        setattr(self.detconb, "input_transform", input_transform)

        transformed_inputs = self.detconb.transform_inputs(inputs)

        if input_transform == "resize_concat":
            assert transformed_inputs.shape[1] == len(in_index)
        elif input_transform == "multiple_select":
            assert len(transformed_inputs) == len(in_index)
        else:
            assert len(transformed_inputs) == 1

    @e2e_pytest_unit
    def test_sample_masked_feats(self) -> None:
        """Test sample_masked_feats function."""
        setattr(self.detconb, "in_index", [0, 1])
        setattr(self.detconb, "input_transform", "resize_concat")
        feats = [torch.randn((1, 1, 4, 4)), torch.randn((1, 1, 2, 2))]
        masks = torch.randint(0, 3, (1, 1, 4, 4), dtype=torch.int64)

        proj, sampled_mask_ids = self.detconb.sample_masked_feats(feats, masks, self.detconb.online_projector)

        assert proj.ndim == 2
        assert proj.shape[0] == reduce(lambda x, y: x * y, feats[0].shape[2:])
        assert proj.shape[1] == sum([feat.shape[1] for feat in feats])
        assert proj.shape[0] == sampled_mask_ids.shape[1]

    @e2e_pytest_unit
    def test_detconb_train_step(self) -> None:
        """Test train_step function wraps forward and _parse_losses."""
        setattr(self.detconb, "in_index", [0, 1])
        setattr(self.detconb, "input_transform", "resize_concat")
        img = torch.randn((1, 2, 3, 4, 4))
        gt_semantic_seg = torch.randint(0, 3, (1, 1, 2, 4, 4), dtype=torch.int64)

        outputs = self.detconb.train_step(
            data_batch=dict(img=img, img_metas={}, gt_semantic_seg=gt_semantic_seg), optimizer=None
        )

        assert "loss" in outputs
        assert "log_vars" in outputs
        assert "num_samples" in outputs


class TestSupConDetConB:
    """Test SupConDetConB."""

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "img,gt_semantic_seg,expected",
        [
            (torch.ones((1, 2, 3, 4, 4), dtype=torch.float32), torch.ones((1, 1, 2, 4, 4), dtype=torch.int64), True),
            (torch.ones((1, 3, 4, 4), dtype=torch.float32), torch.ones((1, 1, 4, 4), dtype=torch.int64), False),
        ],
    )
    def test_forward_train(self, img: torch.Tensor, gt_semantic_seg: torch.Tensor, expected: bool):
        """Test forward_train function."""
        supcon_detconb = SupConDetConB(
            backbone={},
            neck={},
            head={},
            decode_head={},
            downsample=1,
            input_transform="resize_concat",
            in_index=[0, 1],
            loss_cfg=dict(type="DetConLoss"),
            task_adapt=dict(dst_classes=1, src_classes=1),
        )

        results = supcon_detconb(img=img, img_metas=[], gt_semantic_seg=gt_semantic_seg)

        assert ("loss_detcon" in results) == expected



"""Tests for MPA Class-Incremental Learning for object detection with OTX CLI"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import os

import pytest

from otx.api.entities.model_template import parse_model_template
from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    nncf_optimize_testing,
    otx_deploy_openvino_testing,
    otx_eval_deployment_testing,
    otx_eval_openvino_testing,
    otx_eval_testing,
    otx_explain_openvino_testing,
    otx_explain_testing,
    otx_export_testing,
    otx_hpo_testing,
    otx_train_testing,
)

args = {
    "--train-data-roots": "tests/assets/car_tree_bug",
    "--val-data-roots": "tests/assets/car_tree_bug",
    "--test-data-roots": "tests/assets/car_tree_bug",
    "--input": "tests/assets/car_tree_bug/images/train",
    "train_params": [
        "params",
        "--learning_parameters.num_iters",
        "1",
        "--learning_parameters.batch_size",
        "4",
        "--tiling_parameters.enable_tiling",
        "1",
        "--tiling_parameters.enable_adaptive_params",
        "1",
    ],
}

otx_dir = os.getcwd()

default_template = parse_model_template(
    os.path.join("otx/algorithms/detection/configs", "detection", "mobilenetv2_atss", "template.yaml")
)
templates = [default_template]
templates_ids = [default_template.model_template_id]


class TestTilingDetectionCLI:
    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_train_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_export(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_export_testing(template, tmp_dir_path)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_eval_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_eval_openvino_testing(template, tmp_dir_path, otx_dir, args, threshold=1.0)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_explain(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_explain_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_explain_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_explain_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_deploy_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_deploy_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval_deployment(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_eval_deployment_testing(template, tmp_dir_path, otx_dir, args, threshold=1.0)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_hpo(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det/test_hpo"
        otx_hpo_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skip(reason="CVS-98026")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_optimize(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_optimize_testing(template, tmp_dir_path, otx_dir, args)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule

from ..utils import normalize


class IterativeAggregator(nn.Module):
    """Based on: https://github.com/HRNet/Lite-HRNet"""

    def __init__(
        self, in_channels, min_channels=None, conv_cfg=None, norm_cfg=dict(type="BN"), merge_norm=None, use_concat=False
    ):
        super().__init__()

        self.use_concat = use_concat

        num_branches = len(in_channels)
        self.in_channels = in_channels[::-1]

        min_channels = min_channels if min_channels is not None else 0
        assert min_channels >= 0

        out_channels = None
        projects, expanders, fuse_layers = [], [], []
        for i in range(num_branches):
            if not self.use_concat or i == 0:
                fuse_layers.append(None)
            else:
                fuse_layers.append(
                    ConvModule(
                        in_channels=2 * out_channels,
                        out_channels=out_channels,
                        kernel_size=1,
                        stride=1,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        act_cfg=dict(type="ReLU"),
                    )
                )

            if i != num_branches - 1:
                out_channels = max(self.in_channels[i + 1], min_channels)
            else:
                out_channels = max(self.in_channels[i], min_channels)

            projects.append(
                DepthwiseSeparableConvModule(
                    in_channels=max(self.in_channels[i], min_channels),
                    out_channels=out_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=dict(type="ReLU"),
                    dw_act_cfg=None,
                    pw_act_cfg=dict(type="ReLU"),
                )
            )

            if self.in_channels[i] < min_channels:
                expanders.append(
                    ConvModule(
                        in_channels=self.in_channels[i],
                        out_channels=min_channels,
                        kernel_size=1,
                        stride=1,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        act_cfg=dict(type="ReLU"),
                    )
                )
            else:
                expanders.append(None)

        self.projects = nn.ModuleList(projects)
        self.expanders = nn.ModuleList(expanders)
        self.fuse_layers = nn.ModuleList(fuse_layers)

        assert merge_norm in [None, "none", "channel", "spatial"]
        self.merge_norm = merge_norm

    @staticmethod
    def _norm(x, mode=None):
        if mode is None or mode == "none":
            out = x
        elif mode == "channel":
            out = normalize(x, dim=1, p=2)
        else:
            _, c, h, w = x.size()
            y = x.view(-1, c, h * w)
            y = normalize(y, dim=2, p=2)
            out = y.view(-1, c, h, w)

        return out

    def forward(self, x):
        x = x[::-1]

        y_list = []
        last_x = None
        for i, s in enumerate(x):
            if self.expanders[i] is not None:
                s = self.expanders[i](s)

            if last_x is not None:
                last_x = F.interpolate(last_x, size=s.size()[-2:], mode="bilinear", align_corners=True)

                norm_s = self._norm(s, self.merge_norm)
                norm_x = self._norm(last_x, self.merge_norm)

                if self.use_concat:
                    concat_s = torch.cat([norm_s, norm_x], dim=1)
                    s = self.fuse_layers[i](concat_s)
                else:
                    s = norm_s + norm_x

            s = self.projects[i](s)
            last_x = s

            y_list.append(s)

        return y_list[::-1]


class IterativeConcatAggregator(nn.Module):
    def __init__(self, in_channels, min_channels=None, conv_cfg=None, norm_cfg=dict(type="BN"), merge_norm=None):
        super().__init__()

        num_branches = len(in_channels)
        self.in_channels = in_channels[::-1]

        min_channels = min_channels if min_channels is not None else 0
        assert min_channels >= 0

        fuse_layers = [None]
        for i in range(1, num_branches):
            if i == 1:
                num_input_channels = self.in_channels[i - 1] + self.in_channels[i]
            else:
                num_input_channels = max(self.in_channels[i - 1], min_channels) + self.in_channels[i]

            num_out_channels = max(self.in_channels[i], min_channels)

            fuse_layers.append(
                ConvModule(
                    in_channels=num_input_channels,
                    out_channels=num_out_channels,
                    kernel_size=1,
                    stride=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=dict(type="ReLU"),
                )
            )

        self.fuse_layers = nn.ModuleList(fuse_layers)

        assert merge_norm in [None, "none", "channel", "spatial"]
        self.merge_norm = merge_norm

    @staticmethod
    def _norm(x, mode=None):
        if mode is None or mode == "none":
            out = x
        elif mode == "channel":
            out = normalize(x, dim=1, p=2)
        else:
            _, c, h, w = x.size()
            y = x.view(-1, c, h * w)
            y = normalize(y, dim=2, p=2)
            out = y.view(-1, c, h, w)

        return out

    def forward(self, x):
        x = x[::-1]

        y_list = []
        last_x = None
        for i, s in enumerate(x):
            if last_x is not None:
                last_x = F.interpolate(last_x, size=s.size()[-2:], mode="bilinear", align_corners=True)

                norm_s = self._norm(s, self.merge_norm)
                norm_x = self._norm(last_x, self.merge_norm)

                concat_s = torch.cat([norm_s, norm_x], dim=1)
                s = self.fuse_layers[i](concat_s)

            last_x = s
            y_list.append(s)

        return y_list[::-1]



"""This module implements the Coordinate entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from typing import Tuple


class Coordinate:
    """Represents a 2D-coordinate with an x-position and a y-position.

    NB most coordinates are normalized (between 0.0 and 1.0)

    Args:
        x (float): x-coordinate
        y (float): y-coordinate
    """

    __slots__ = ["x", "y"]

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __repr__(self):
        """String representation of the coordinate. Returns the x and y coordinates."""
        return f"Coordinate(x={self.x}, y={self.y})"

    def __eq__(self, other):
        """Returns True if the coordinates are equal."""
        return self.x == other.x and self.y == other.y

    def __hash__(self):
        """Returns the hash of the coordinate."""
        return hash(str(self))

    def as_tuple(self) -> Tuple[float, float]:
        """Convert the coordinates to a pair (x,y)."""
        return self.x, self.y

    def as_int_tuple(self) -> Tuple[int, int]:
        """Convert the coordinates to a pair of integer coordinates (x,y)."""
        return int(self.x), int(self.y)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcv.runner import force_fp32
from mmdet.core import bbox_overlaps, distance2bbox, reduce_mean
from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.vfnet_head import VFNetHead

from otx.mpa.modules.models.heads.cross_dataset_detector_head import (
    CrossDatasetDetectorHead,
)
from otx.mpa.modules.models.losses.cross_focal_loss import CrossSigmoidFocalLoss


@HEADS.register_module()
class CustomVFNetHead(CrossDatasetDetectorHead, VFNetHead):
    def __init__(self, *args, bg_loss_weight=-1.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.bg_loss_weight = bg_loss_weight

    @force_fp32(apply_to=("cls_scores", "bbox_preds", "bbox_preds_refine"))
    def loss(self, cls_scores, bbox_preds, bbox_preds_refine, gt_bboxes, gt_labels, img_metas, gt_bboxes_ignore=None):
        """Compute loss of the head.

        Args:
            cls_scores (list[Tensor]): Box iou-aware scores for each scale
                level, each is a 4D-tensor, the channel number is
                num_points * num_classes.
            bbox_preds (list[Tensor]): Box offsets for each
                scale level, each is a 4D-tensor, the channel number is
                num_points * 4.
            bbox_preds_refine (list[Tensor]): Refined Box offsets for
                each scale level, each is a 4D-tensor, the channel
                number is num_points * 4.
            gt_bboxes (list[Tensor]): Ground truth bboxes for each image with
                shape (num_gts, 4) in [tl_x, tl_y, br_x, br_y] format.
            gt_labels (list[Tensor]): class indices corresponding to each box
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            gt_bboxes_ignore (None | list[Tensor]): specify which bounding
                boxes can be ignored when computing the loss.
                Default: None.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        assert len(cls_scores) == len(bbox_preds) == len(bbox_preds_refine)
        featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
        all_level_points = self.get_points(featmap_sizes, bbox_preds[0].dtype, bbox_preds[0].device)
        labels, label_weights, bbox_targets, bbox_weights, valid_label_mask = self.get_targets(
            cls_scores, all_level_points, gt_bboxes, gt_labels, img_metas, gt_bboxes_ignore
        )

        num_imgs = cls_scores[0].size(0)
        # flatten cls_scores, bbox_preds and bbox_preds_refine
        flatten_cls_scores = [
            cls_score.permute(0, 2, 3, 1).reshape(-1, self.cls_out_channels).contiguous() for cls_score in cls_scores
        ]
        flatten_bbox_preds = [bbox_pred.permute(0, 2, 3, 1).reshape(-1, 4).contiguous() for bbox_pred in bbox_preds]
        flatten_bbox_preds_refine = [
            bbox_pred_refine.permute(0, 2, 3, 1).reshape(-1, 4).contiguous() for bbox_pred_refine in bbox_preds_refine
        ]
        flatten_cls_scores = torch.cat(flatten_cls_scores)
        flatten_bbox_preds = torch.cat(flatten_bbox_preds)
        flatten_bbox_preds_refine = torch.cat(flatten_bbox_preds_refine)
        flatten_labels = torch.cat(labels)
        flatten_bbox_targets = torch.cat(bbox_targets)

        valid_label_mask = [
            valid_mask.reshape(-1, self.cls_out_channels).contiguous() for valid_mask in valid_label_mask
        ]
        flatten_valid_label_mask = torch.cat(valid_label_mask)
        # repeat points to align with bbox_preds
        flatten_points = torch.cat([points.repeat(num_imgs, 1) for points in all_level_points])

        # FG cat_id: [0, num_classes - 1], BG cat_id: num_classes
        bg_class_ind = self.num_classes
        pos_inds = torch.where(((flatten_labels >= 0) & (flatten_labels < bg_class_ind)) > 0)[0]
        num_pos = len(pos_inds)

        pos_bbox_preds = flatten_bbox_preds[pos_inds]
        pos_bbox_preds_refine = flatten_bbox_preds_refine[pos_inds]
        pos_labels = flatten_labels[pos_inds]

        # sync num_pos across all gpus
        if self.sync_num_pos:
            num_pos_avg_per_gpu = reduce_mean(pos_inds.new_tensor(num_pos).float()).item()
            num_pos_avg_per_gpu = max(num_pos_avg_per_gpu, 1.0)
        else:
            num_pos_avg_per_gpu = num_pos
        if num_pos > 0:
            pos_bbox_targets = flatten_bbox_targets[pos_inds]
            pos_points = flatten_points[pos_inds]

            pos_decoded_bbox_preds = distance2bbox(pos_points, pos_bbox_preds)
            pos_decoded_target_preds = distance2bbox(pos_points, pos_bbox_targets)
            iou_targets_ini = bbox_overlaps(
                pos_decoded_bbox_preds, pos_decoded_target_preds.detach(), is_aligned=True
            ).clamp(min=1e-6)
            bbox_weights_ini = iou_targets_ini.clone().detach()
            iou_targets_ini_avg_per_gpu = reduce_mean(bbox_weights_ini.sum()).item()
            bbox_avg_factor_ini = max(iou_targets_ini_avg_per_gpu, 1.0)
            loss_bbox = self.loss_bbox(
                pos_decoded_bbox_preds,
                pos_decoded_target_preds.detach(),
                weight=bbox_weights_ini,
                avg_factor=bbox_avg_factor_ini,
            )

            pos_decoded_bbox_preds_refine = distance2bbox(pos_points, pos_bbox_preds_refine)
            iou_targets_rf = bbox_overlaps(
                pos_decoded_bbox_preds_refine, pos_decoded_target_preds.detach(), is_aligned=True
            ).clamp(min=1e-6)
            bbox_weights_rf = iou_targets_rf.clone().detach()
            iou_targets_rf_avg_per_gpu = reduce_mean(bbox_weights_rf.sum()).item()
            bbox_avg_factor_rf = max(iou_targets_rf_avg_per_gpu, 1.0)
            loss_bbox_refine = self.loss_bbox_refine(
                pos_decoded_bbox_preds_refine,
                pos_decoded_target_preds.detach(),
                weight=bbox_weights_rf,
                avg_factor=bbox_avg_factor_rf,
            )

            # build IoU-aware cls_score targets
            if self.use_vfl:
                pos_ious = iou_targets_rf.clone().detach()
                cls_iou_targets = torch.zeros_like(flatten_cls_scores)
                cls_iou_targets[pos_inds, pos_labels] = pos_ious
        else:
            loss_bbox = pos_bbox_preds.sum() * 0
            loss_bbox_refine = pos_bbox_preds_refine.sum() * 0
            if self.use_vfl:
                cls_iou_targets = torch.zeros_like(flatten_cls_scores)
        # Re-weigting BG loss
        if self.bg_loss_weight >= 0.0:
            neg_indices = flatten_labels == self.num_classes
            label_weights[neg_indices] = self.bg_loss_weight

        if self.use_vfl:
            if label_weights is not None:
                label_weights = label_weights.unsqueeze(-1)
            if isinstance(self.loss_cls, CrossSigmoidFocalLoss):
                loss_cls = self.loss_cls(
                    flatten_cls_scores,
                    cls_iou_targets,
                    weight=label_weights,
                    avg_factor=num_pos_avg_per_gpu,
                    use_vfl=self.use_vfl,
                    valid_label_mask=flatten_valid_label_mask,
                )
            else:
                loss_cls = self.loss_cls(
                    flatten_cls_scores, cls_iou_targets, weight=label_weights, avg_factor=num_pos_avg_per_gpu
                )
        else:
            loss_cls = self.loss_cls(
                flatten_cls_scores, flatten_labels, weight=label_weights, avg_factor=num_pos_avg_per_gpu
            )

        return dict(loss_cls=loss_cls, loss_bbox=loss_bbox, loss_bbox_rf=loss_bbox_refine)

    def get_targets(self, cls_scores, mlvl_points, gt_bboxes, gt_labels, img_metas, gt_bboxes_ignore):
        """A wrapper for computing ATSS and FCOS targets for points in multiple
        images.

        Args:
            cls_scores (list[Tensor]): Box iou-aware scores for each scale
                level with shape (N, num_points * num_classes, H, W).
            mlvl_points (list[Tensor]): Points of each fpn level, each has
                shape (num_points, 2).
            gt_bboxes (list[Tensor]): Ground truth bboxes of each image,
                each has shape (num_gt, 4).
            gt_labels (list[Tensor]): Ground truth labels of each box,
                each has shape (num_gt,).
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            gt_bboxes_ignore (None | Tensor): Ground truth bboxes to be
                ignored, shape (num_ignored_gts, 4).

        Returns:
            tuple:
                labels_list (list[Tensor]): Labels of each level.
                label_weights (Tensor/None): Label weights of all levels.
                bbox_targets_list (list[Tensor]): Regression targets of each
                    level, (l, t, r, b).
                bbox_weights (Tensor/None): Bbox weights of all levels.
        """
        if self.use_atss:
            return self.vfnet_to_atss_targets(
                cls_scores, mlvl_points, gt_bboxes, gt_labels, img_metas, gt_bboxes_ignore
            )
        else:
            self.norm_on_bbox = False
            return self.get_fcos_targets(mlvl_points, gt_bboxes, gt_labels, img_metas)



"""Tests for Class-Incremental Learning for object detection with OTX CLI"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import copy
import os

import pytest
import torch

from otx.api.entities.model_template import parse_model_template
from otx.cli.registry import Registry
from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    get_template_dir,
    nncf_eval_openvino_testing,
    nncf_eval_testing,
    nncf_export_testing,
    nncf_optimize_testing,
    nncf_validate_fq_testing,
    otx_demo_deployment_testing,
    otx_demo_openvino_testing,
    otx_demo_testing,
    otx_deploy_openvino_testing,
    otx_eval_deployment_testing,
    otx_eval_openvino_testing,
    otx_eval_testing,
    otx_explain_openvino_testing,
    otx_explain_testing,
    otx_export_testing,
    otx_hpo_testing,
    otx_resume_testing,
    otx_train_testing,
    pot_eval_testing,
    pot_optimize_testing,
    pot_validate_fq_testing,
)

# Pre-train w/ 'person' class ##TODO: Currently, it is closed to sample test. need to change other sample
args0 = {
    "--train-data-roots": "tests/assets/car_tree_bug",
    "--val-data-roots": "tests/assets/car_tree_bug",
    "--test-data-roots": "tests/assets/car_tree_bug",
    "--input": "tests/assets/car_tree_bug/images/train",
    "train_params": ["params", "--learning_parameters.num_iters", "4", "--learning_parameters.batch_size", "4"],
}

# Class-Incremental learning w/ 'vehicle', 'person', 'non-vehicle' classes
args = {
    "--train-data-roots": "tests/assets/car_tree_bug",
    "--val-data-roots": "tests/assets/car_tree_bug",
    "--test-data-roots": "tests/assets/car_tree_bug",
    "--input": "tests/assets/car_tree_bug/images/train",
    "train_params": ["params", "--learning_parameters.num_iters", "2", "--learning_parameters.batch_size", "4"],
}

args_semisl = {
    "--train-data-roots": "tests/assets/car_tree_bug",
    "--val-data-roots": "tests/assets/car_tree_bug",
    "--test-data-roots": "tests/assets/car_tree_bug",
    "--unlabeled-data-roots": "tests/assets/car_tree_bug",
    "--input": "tests/assets/car_tree_bug/images/train",
    "train_params": [
        "params",
        "--learning_parameters.num_iters",
        "2",
        "--learning_parameters.batch_size",
        "4",
        "--algo_backend.train_type",
        "SEMISUPERVISED",
    ],
}

# Training params for resume, num_iters*2
resume_params = [
    "params",
    "--learning_parameters.num_iters",
    "8",
    "--learning_parameters.batch_size",
    "4",
]

otx_dir = os.getcwd()

MULTI_GPU_UNAVAILABLE = torch.cuda.device_count() <= 1
TT_STABILITY_TESTS = os.environ.get("TT_STABILITY_TESTS", False)
if TT_STABILITY_TESTS:
    default_template = parse_model_template(
        os.path.join("otx/algorithms/detection/configs", "detection", "mobilenetv2_atss", "template.yaml")
    )
    templates = [default_template] * 100
    templates_ids = [template.model_template_id + f"-{i+1}" for i, template in enumerate(templates)]
else:
    templates = Registry("otx/algorithms/detection").filter(task_type="DETECTION").templates
    templates_ids = [template.model_template_id for template in templates]


class TestToolsMPADetection:
    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_train_testing(template, tmp_dir_path, otx_dir, args0)
        template_work_dir = get_template_dir(template, tmp_dir_path)
        args1 = copy.deepcopy(args)
        args1["--load-weights"] = f"{template_work_dir}/trained_{template.model_template_id}/weights.pth"
        otx_train_testing(template, tmp_dir_path, otx_dir, args1)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_resume(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection/test_resume"
        otx_resume_testing(template, tmp_dir_path, otx_dir, args0)
        template_work_dir = get_template_dir(template, tmp_dir_path)
        args1 = copy.deepcopy(args0)
        args1["train_params"] = resume_params
        args1["--resume-from"] = f"{template_work_dir}/trained_for_resume_{template.model_template_id}/weights.pth"
        otx_resume_testing(template, tmp_dir_path, otx_dir, args1)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_export(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_export_testing(template, tmp_dir_path)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_eval_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_eval_openvino_testing(template, tmp_dir_path, otx_dir, args, threshold=0.2)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_explain(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_explain_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_explain_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_explain_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_demo_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_demo_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_deploy_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_deploy_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval_deployment(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_eval_deployment_testing(template, tmp_dir_path, otx_dir, args, threshold=0.0)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo_deployment(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        otx_demo_deployment_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_hpo(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection/test_hpo"
        otx_hpo_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_optimize(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_optimize_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_export(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_export_testing(template, tmp_dir_path)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_validate_fq(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_validate_fq_testing(template, tmp_dir_path, otx_dir, "detection", type(self).__name__)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_eval_testing(template, tmp_dir_path, otx_dir, args, threshold=0.001)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_eval_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_eval_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_optimize(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        pot_optimize_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_validate_fq(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        pot_validate_fq_testing(template, tmp_dir_path, otx_dir, "detection", type(self).__name__)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection"
        pot_eval_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skip(reason="CVS-101246 Multi-GPU tests are stuck while CI is running")
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.skipif(MULTI_GPU_UNAVAILABLE, reason="The number of gpu is insufficient")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_multi_gpu_train(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection/test_multi_gpu"
        args1 = copy.deepcopy(args)
        args1["--gpus"] = "0,1"
        otx_train_testing(template, tmp_dir_path, otx_dir, args1)


class TestToolsMPASemiSLDetection:
    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection/test_semisl"
        otx_train_testing(template, tmp_dir_path, otx_dir, args_semisl)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "detection/test_semisl"
        otx_eval_testing(template, tmp_dir_path, otx_dir, args)



"""Simple monkey patch helper."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=unnecessary-dunder-call,invalid-name

import ctypes
import importlib
import inspect
from collections import OrderedDict
from functools import partial, partialmethod
from typing import Callable


class Patcher:
    """Simple monkey patch helper."""

    def __init__(self):
        self._patched = OrderedDict()

    def patch(  # noqa: C901
        self,
        obj_cls,
        wrapper: Callable,
        *,
        force: bool = True,
    ):
        """Do monkey patch."""
        if isinstance(obj_cls, (tuple, list)):
            assert len(obj_cls) == 2
            obj_cls, fn_name = obj_cls
            assert getattr(obj_cls, fn_name)
        else:
            obj_cls, fn_name = self.import_obj(obj_cls)

        # wrap only if function does exist
        n_args = len(inspect.getfullargspec(obj_cls.__getattribute__)[0])
        if n_args == 1:
            try:
                fn = obj_cls.__getattribute__(fn_name)
            except AttributeError:
                return
            self._patch_module_fn(obj_cls, fn_name, fn, wrapper, force)
        else:
            if inspect.isclass(obj_cls):
                try:
                    fn = obj_cls.__getattribute__(obj_cls, fn_name)  # type: ignore
                except AttributeError:
                    return
                self._patch_class_fn(obj_cls, fn_name, fn, wrapper, force)
            else:
                try:
                    fn = obj_cls.__getattribute__(fn_name)
                except AttributeError:
                    return
                self._patch_instance_fn(obj_cls, fn_name, fn, wrapper, force)

    def unpatch(self, obj_cls=None, depth=0):
        """Undo monkey patch."""

        def _unpatch(obj, fn_name, key, depth):
            if depth == 0:
                depth = len(self._patched[key])
            keep = len(self._patched[key]) - depth
            origin_fn = self._patched[key].pop(-depth)[0]
            while self._patched[key] and len(self._patched[key]) > keep:
                self._patched[key].pop()
            if not self._patched[key]:
                self._patched.pop(key)

            if isinstance(obj, int):
                obj = ctypes.cast(obj, ctypes.py_object).value
            setattr(obj, fn_name, origin_fn)

        if obj_cls is not None:
            obj_cls, fn_name = self.import_obj(obj_cls)
            n_args = len(inspect.getfullargspec(obj_cls.__getattribute__)[0])
            if n_args == 1:
                key = (obj_cls.__name__, fn_name)
            else:
                if inspect.isclass(obj_cls):
                    obj_cls_path = obj_cls.__module__ + "." + obj_cls.__name__
                    key = (obj_cls_path, fn_name)
                else:
                    key = (id(obj_cls), fn_name)
            _unpatch(obj_cls, fn_name, key, depth)
            return

        for key in list(self._patched.keys()):
            obj, fn_name = key
            if isinstance(obj, int):
                obj = ctypes.cast(obj, ctypes.py_object).value
            else:
                obj, fn_name = self.import_obj(".".join([obj, fn_name]))
            _unpatch(obj, fn_name, key, depth)

    def import_obj(self, obj_cls):  # noqa: C901
        """Object import helper."""
        if isinstance(obj_cls, str):
            fn_name = obj_cls.split(".")[-1]
            obj_cls = ".".join(obj_cls.split(".")[:-1])
        else:
            if "_partialmethod" in obj_cls.__dict__:
                while "_partialmethod" in obj_cls.__dict__:
                    obj_cls = obj_cls._partialmethod.keywords["__fn"]  # pylint: disable=protected-access
            while isinstance(obj_cls, (partial, partialmethod)):
                obj_cls = obj_cls.keywords["__fn"]

            if inspect.ismodule(obj_cls):
                fn = obj_cls.keywords["__fn"]
                fn_name = fn.__name__
                obj_cls = fn = obj_cls.keywords["__obj_cls"]
            elif inspect.ismethod(obj_cls):
                fn_name = obj_cls.__name__
                obj_cls = obj_cls.__self__
            elif isinstance(obj_cls, (staticmethod, classmethod)):
                obj_cls = obj_cls.__func__
                fn_name = obj_cls.__name__
                obj_cls = ".".join([obj_cls.__module__] + obj_cls.__qualname__.split(".")[:-1])
            else:
                fn_name = obj_cls.__name__
                obj_cls = ".".join([obj_cls.__module__] + obj_cls.__qualname__.split(".")[:-1])

        if isinstance(obj_cls, str):
            try:
                obj_cls = importlib.import_module(obj_cls)
            except ModuleNotFoundError:
                module = ".".join(obj_cls.split(".")[:-1])
                obj_cls = obj_cls.split(".")[-1]
                obj_cls = getattr(importlib.import_module(module), obj_cls)
        return obj_cls, fn_name

    def _patch_module_fn(self, obj_cls, fn_name, fn, wrapper, force):
        def helper(*args, **kwargs):  # type: ignore
            obj_cls = kwargs.pop("__obj_cls")
            fn = kwargs.pop("__fn")
            wrapper = kwargs.pop("__wrapper")
            return wrapper(obj_cls, fn, *args, **kwargs)

        assert len(inspect.getfullargspec(obj_cls.__getattribute__)[0]) == 1
        obj_cls_path = obj_cls.__name__
        key = (obj_cls_path, fn_name)
        fn_ = self._initialize(key, force)
        if fn_ is not None:
            fn = fn_
        setattr(obj_cls, fn_name, partial(helper, __wrapper=wrapper, __fn=fn, __obj_cls=obj_cls))
        self._patched[key].append((fn, wrapper))

    def _patch_class_fn(self, obj_cls, fn_name, fn, wrapper, force):

        if isinstance(fn, (staticmethod, classmethod)):

            def helper(*args, **kwargs):  # type: ignore
                wrapper = kwargs.pop("__wrapper")
                fn = kwargs.pop("__fn")
                obj_cls = kwargs.pop("__obj_cls")
                if isinstance(args[0], obj_cls):
                    return wrapper(args[0], fn.__get__(args[0]), *args[1:], **kwargs)
                return wrapper(obj_cls, fn.__get__(obj_cls), *args, **kwargs)

        elif isinstance(fn, type(all.__call__)):

            def helper(self, *args, **kwargs):  # type: ignore
                kwargs.pop("__obj_cls")
                wrapper = kwargs.pop("__wrapper")
                fn = kwargs.pop("__fn")
                return wrapper(self, fn, *args, **kwargs)

        else:

            def helper(self, *args, **kwargs):  # type: ignore
                kwargs.pop("__obj_cls")
                wrapper = kwargs.pop("__wrapper")
                fn = kwargs.pop("__fn")
                return wrapper(self, fn.__get__(self), *args, **kwargs)

        assert len(inspect.getfullargspec(obj_cls.__getattribute__)[0]) == 2
        obj_cls_path = obj_cls.__module__ + "." + obj_cls.__name__
        key = (obj_cls_path, fn_name)
        fn_ = self._initialize(key, force)
        if fn_ is not None:
            fn = fn_
        setattr(
            obj_cls,
            fn_name,
            partialmethod(helper, __wrapper=wrapper, __fn=fn, __obj_cls=obj_cls),
        )
        self._patched[key].append((fn, wrapper))

    def _patch_instance_fn(self, obj_cls, fn_name, fn, wrapper, force):
        def helper(ctx, *args, **kwargs):  # type: ignore
            fn = kwargs.pop("__fn")
            wrapper = kwargs.pop("__wrapper")
            return wrapper(ctx, fn, *args, **kwargs)

        assert len(inspect.getfullargspec(obj_cls.__getattribute__)[0]) == 2
        obj_cls_path = id(obj_cls)
        key = (obj_cls_path, fn_name)
        fn_ = self._initialize(key, force)
        if fn_ is not None:
            fn = fn_
        setattr(obj_cls, fn_name, partialmethod(helper, __wrapper=wrapper, __fn=fn).__get__(obj_cls))
        self._patched[key].append((fn, wrapper))

    def _initialize(self, key, force):
        fn = None
        if key not in self._patched:
            self._patched[key] = []
        if force:
            while self._patched[key]:
                fn, *_ = self._patched[key].pop()
        return fn



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .activations import *
from .arithmetics import *
from .builder import OPS
from .convolutions import *
from .generation import *
from .image_processings import *
from .infrastructures import *
from .matmuls import *
from .movements import *
from .normalizations import *
from .object_detections import *

# flake8: noqa
from .op import *
from .poolings import *
from .reductions import *
from .shape_manipulations import *
from .sorting_maximization import *
from .type_conversions import *



"""This module contains the f-measure performance provider class."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import logging
from typing import Dict, List, Optional, Tuple

import numpy as np

from otx.api.entities.annotation import Annotation
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    ColorPalette,
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    MetricsGroup,
    Performance,
    ScoreMetric,
    TextChartInfo,
    TextMetricsGroup,
    VisualizationType,
)
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.evaluation.performance_provider_interface import (
    IPerformanceProvider,
)
from otx.api.utils.shape_factory import ShapeFactory

logger = logging.getLogger(__name__)
ALL_CLASSES_NAME = "All Classes"


def intersection_box(
    box1: Tuple[float, float, float, float, str, float], box2: Tuple[float, float, float, float, str, float]
) -> Tuple[float, float, float, float]:
    """Calculate the intersection box of two bounding boxes.

    Args:
        box1 (Tuple[float, float, float, float, str, float]): (x1, y1, x2, y2, class, score)
        box2 (Tuple[float, float, float, float, str, float]): (x1, y1, x2, y2, class, score)

    Returns:
        Tuple[float, float, float, float]: (x_left, x_right, y_bottom, y_top)
    """
    x_left = max(box1[0], box2[0])
    y_top = max(box1[1], box2[1])
    x_right = min(box1[2], box2[2])
    y_bottom = min(box1[3], box2[3])
    return (x_left, x_right, y_bottom, y_top)


def bounding_box_intersection_over_union(
    box1: Tuple[float, float, float, float, str, float], box2: Tuple[float, float, float, float, str, float]
) -> float:
    """Calculate the Intersection over Union (IoU) of two bounding boxes.

    Args:
        box1 (Tuple[float, float, float, float, str, float]): (x1, y1, x2, y2, class, score)
        box2 (Tuple[float, float, float, float, str, float]): (x1, y1, x2, y2, class, score)

    Raises:
        ValueError: In case the IoU is outside of [0.0, 1.0]

    Returns:
        float: Intersection-over-union of box1 and box2.
    """

    x_left, x_right, y_bottom, y_top = intersection_box(box1, box2)

    if x_right <= x_left or y_bottom <= y_top:
        iou = 0.0
    else:
        intersection_area = (x_right - x_left) * (y_bottom - y_top)
        bb1_area = (box1[2] - box1[0]) * (box1[3] - box1[1])
        bb2_area = (box2[2] - box2[0]) * (box2[3] - box2[1])
        union_area = float(bb1_area + bb2_area - intersection_area)
        if union_area == 0:
            iou = 0.0
        else:
            iou = intersection_area / union_area
    if iou < 0.0 or iou > 1.0:
        raise ValueError(f"intersection over union should be in range [0,1], actual={iou}")
    return iou


def get_iou_matrix(
    ground_truth: List[Tuple[float, float, float, float, str, float]],
    predicted: List[Tuple[float, float, float, float, str, float]],
) -> np.ndarray:
    """Constructs an iou matrix of shape [num_ground_truth_boxes, num_predicted_boxes].

    Each cell(x,y) in the iou matrix contains the intersection over union of ground truth box(x) and predicted box(y)
    An iou matrix corresponds to a single image

    Args:
        ground_truth (List[Tuple[float, float, float, float, str, float]]): List of ground truth boxes.
            Each box is a list of (x,y) coordinates and a label.
            a box: [x1: float, y1, x2, y2, class: str, score: float]
            boxes_per_image: [box1, box2, …]
            boxes1: [boxes_per_image_1, boxes_per_image_2, boxes_per_image_3, …]
        predicted (List[Tuple[float, float, float, float, str, float]]): List of predicted boxes.
            Each box is a list of (x,y) coordinates and a label.
            a box: [x1: float, y1, x2, y2, class: str, score: float]
            boxes_per_image: [box1, box2, …]
            boxes2: [boxes_per_image_1, boxes_per_image_2, boxes_per_image_3, …]

    Returns:
        np.ndarray: IoU matrix of shape [ground_truth_boxes, predicted_boxes]
    """
    matrix = np.array(
        [[bounding_box_intersection_over_union(gts, preds) for preds in predicted] for gts in ground_truth]
    )
    return matrix


def get_n_false_negatives(iou_matrix: np.ndarray, iou_threshold: float) -> int:
    """Get the number of false negatives inside the IoU matrix for a given threshold.

    The first loop accounts for all the ground truth boxes which do not have a high enough iou with any predicted
    box (they go undetected)
    The second loop accounts for the much rarer case where two ground truth boxes are detected by the same predicted
    box. The principle is that each ground truth box requires a unique prediction box

    Args:
        iou_matrix (np.ndarray): IoU matrix of shape [ground_truth_boxes, predicted_boxes]
        iou_threshold (float): IoU threshold to use for the false negatives.

    Returns:
        int: Number of false negatives
    """
    n_false_negatives = 0
    for row in iou_matrix:
        if max(row) < iou_threshold:
            n_false_negatives += 1
    for column in np.rot90(iou_matrix):
        indices = np.where(column > iou_threshold)
        n_false_negatives += max(len(indices[0]) - 1, 0)
    return n_false_negatives


class _Metrics:
    """This class collects the metrics related to detection.

    Args:
        f_measure (float): F-measure of the model.
        precision (float): Precision of the model.
        recall (float): Recall of the model.
    """

    def __init__(self, f_measure: float, precision: float, recall: float):
        self.f_measure = f_measure
        self.precision = precision
        self.recall = recall


class _ResultCounters:
    """This class collects the number of prediction, TP and FN.

    Args:
        n_false_negatives (int): Number of false negatives.
        n_true (int): Number of true positives.
        n_predictions (int): Number of predictions.
    """

    def __init__(self, n_false_negatives: int, n_true: int, n_predicted: int):
        self.n_false_negatives = n_false_negatives
        self.n_true = n_true
        self.n_predicted = n_predicted

    def calculate_f_measure(self) -> _Metrics:
        """Calculates and returns precision, recall, and f-measure.

        Returns:
            _Metrics: _Metrics object with Precision, recall, and f-measure.
        """
        n_true_positives = self.n_true - self.n_false_negatives

        if self.n_predicted == 0:
            precision = 1.0
            recall = 0.0
        elif self.n_true == 0:
            precision = 0.0
            recall = 1.0
        else:
            precision = n_true_positives / self.n_predicted
            recall = n_true_positives / self.n_true

        f_measure = (2 * precision * recall) / (precision + recall + np.finfo(float).eps)
        return _Metrics(f_measure, precision, recall)


class _AggregatedResults:
    """This class collects the aggregated results for F-measure.

    The result contains:
        - f_measure_curve
        - precision_curve
        - recall_curve
        - all_classes_f_measure_curve
        - best_f_measure
        - best_threshold

    Args:
        classes (List[str]): List of classes.
    """

    def __init__(self, classes: List[str]):
        self.f_measure_curve: Dict[str, List[float]] = {class_name: [] for class_name in classes}
        self.precision_curve: Dict[str, List[float]] = {class_name: [] for class_name in classes}
        self.recall_curve: Dict[str, List[float]] = {class_name: [] for class_name in classes}
        self.all_classes_f_measure_curve: List[float] = []
        self.best_f_measure: float = 0.0
        self.best_threshold: float = 0.0


class _OverallResults:
    """This class collects the overall results that is computed by the F-measure performance provider.

    Args:
        per_confidence (_AggregatedResults): _AggregatedResults object for each confidence level.
        per_nms (Optional[_AggregatedResults]): _AggregatedResults object for each NMS threshold.
        best_f_measure_per_class (Dict[str, float]): Best f-measure per class.
        best_f_measure (float): Best f-measure.
    """

    def __init__(
        self,
        per_confidence: _AggregatedResults,
        per_nms: Optional[_AggregatedResults],
        best_f_measure_per_class: Dict[str, float],
        best_f_measure: float,
    ):
        self.per_confidence = per_confidence
        self.per_nms = per_nms
        self.best_f_measure_per_class = best_f_measure_per_class
        self.best_f_measure = best_f_measure


class _FMeasureCalculator:
    """This class contains the functions to calculate FMeasure.

    Args:
        ground_truth_boxes_per_image (List[List[Tuple[float, float, float, float, str, float]]]):
                a box: [x1: float, y1, x2, y2, class: str, score: float]
                boxes_per_image: [box1, box2, …]
                ground_truth_boxes_per_image: [boxes_per_image_1, boxes_per_image_2, boxes_per_image_3, …]
        prediction_boxes_per_image (List[List[Tuple[float, float, float, float, str, float]]]):
                a box: [x1: float, y1, x2, y2, class: str, score: float]
                boxes_per_image: [box1, box2, …]
                predicted_boxes_per_image: [boxes_per_image_1, boxes_per_image_2, boxes_per_image_3, …]
    """

    def __init__(
        self,
        ground_truth_boxes_per_image: List[List[Tuple[float, float, float, float, str, float]]],
        prediction_boxes_per_image: List[List[Tuple[float, float, float, float, str, float]]],
    ):
        self.ground_truth_boxes_per_image = ground_truth_boxes_per_image
        self.prediction_boxes_per_image = prediction_boxes_per_image
        self.confidence_range = [0.025, 1.0, 0.025]
        self.nms_range = [0.1, 1, 0.05]
        self.default_confidence_threshold = 0.35

    def evaluate_detections(
        self,
        classes: List[str],
        iou_threshold: float = 0.5,
        result_based_nms_threshold: bool = False,
        cross_class_nms: bool = False,
    ) -> _OverallResults:
        """Evaluates detections by computing f_measures across multiple confidence thresholds and iou thresholds.

        By default, this function evaluates 39 confidence thresholds, finds the best confidence threshold and appends
        it to the result Dict
        Each one of the (default 39+20) pairs of confidence and nms thresholds is used to evaluate the f-measure for
        each class, then the intermediate metrics are summed across classes to compute an all_classes f_measure.
        Finally, the best results across all evaluations are appended to the result dictionary along with the thresholds
        used to achieve them.

        Args:
            classes (List[str]): Names of classes to be evaluated.
            iou_threshold (float): IOU threshold. Defaults to 0.5.
            result_based_nms_threshold (bool): Boolean that determines whether multiple nms threshold are examined.
                Defaults to False.
            cross_class_nms (bool): Set to True to perform NMS between boxes with different classes. Defaults to False.

        Returns:
            _OverallResults: _OverallResults object with the result statistics (e.g F-measure).
        """

        best_f_measure_per_class = {}

        results_per_confidence = self.get_results_per_confidence(
            classes=classes,
            confidence_range=self.confidence_range,
            iou_threshold=iou_threshold,
        )

        best_f_measure = results_per_confidence.best_f_measure

        for class_name in classes:
            best_f_measure_per_class[class_name] = max(results_per_confidence.f_measure_curve[class_name])

        results_per_nms: Optional[_AggregatedResults] = None

        if result_based_nms_threshold:
            results_per_nms = self.get_results_per_nms(
                classes=classes,
                iou_threshold=iou_threshold,
                min_f_measure=results_per_confidence.best_f_measure,
                cross_class_nms=cross_class_nms,
            )

            for class_name in classes:
                best_f_measure_per_class[class_name] = max(results_per_nms.f_measure_curve[class_name])

        result = _OverallResults(
            results_per_confidence,
            results_per_nms,
            best_f_measure_per_class,
            best_f_measure,
        )

        return result

    def get_results_per_confidence(
        self, classes: List[str], confidence_range: List[float], iou_threshold: float
    ) -> _AggregatedResults:
        """Returns the results for confidence threshold in range confidence_range.

        Varies confidence based on confidence_range, the results are appended in a dictionary and returned, it also
        returns the best f_measure found and the confidence threshold used to get said f_measure

        Args:
            classes (List[str]): Names of classes to be evaluated.
            confidence_range (List[float]): List of confidence thresholds to be evaluated.
            iou_threshold (float): IoU threshold to use for false negatives.

        Returns:
            _AggregatedResults: _AggregatedResults object with the result statistics (e.g F-measure).
        """
        result = _AggregatedResults(classes)
        result.best_threshold = 0.1

        for confidence_threshold in np.arange(*confidence_range):
            result_point = self.evaluate_classes(
                classes=classes.copy(),
                iou_threshold=iou_threshold,
                confidence_threshold=confidence_threshold,
            )
            all_classes_f_measure = result_point[ALL_CLASSES_NAME].f_measure
            result.all_classes_f_measure_curve.append(all_classes_f_measure)

            for class_name in classes:
                result.f_measure_curve[class_name].append(result_point[class_name].f_measure)
                result.precision_curve[class_name].append(result_point[class_name].precision)
                result.recall_curve[class_name].append(result_point[class_name].recall)
            if all_classes_f_measure > result.best_f_measure:
                result.best_f_measure = all_classes_f_measure
                result.best_threshold = confidence_threshold
        return result

    def get_results_per_nms(
        self,
        classes: List[str],
        iou_threshold: float,
        min_f_measure: float,
        cross_class_nms: bool = False,
    ) -> _AggregatedResults:
        """Returns results for nms threshold in range nms_range.

        First, we calculate the critical nms of each box, meaning the nms_threshold
        that would cause it to be disappear
        This is an expensive O(n**2) operation, however, doing this makes filtering for every single nms_threshold much
        faster at O(n)

        Args:
            classes (List[str]): List of classes
            iou_threshold (float): IoU threshold
            min_f_measure (float): the minimum F-measure required to select a NMS threshold
            cross_class_nms (bool): set to True to perform NMS between boxes with different classes. Defaults to False.

        Returns:
            _AggregatedResults: Object containing the results for each NMS threshold value
        """
        result = _AggregatedResults(classes)
        result.best_f_measure = min_f_measure
        result.best_threshold = 0.5

        critical_nms_per_image = self.__get_critical_nms(self.prediction_boxes_per_image, cross_class_nms)

        for nms_threshold in np.arange(*self.nms_range):
            predicted_boxes_per_image_per_nms = self.__filter_nms(
                self.prediction_boxes_per_image, critical_nms_per_image, nms_threshold
            )
            boxes_pair_for_nms = _FMeasureCalculator(
                self.ground_truth_boxes_per_image, predicted_boxes_per_image_per_nms
            )
            result_point = boxes_pair_for_nms.evaluate_classes(
                classes=classes.copy(),
                iou_threshold=iou_threshold,
                confidence_threshold=self.default_confidence_threshold,
            )
            all_classes_f_measure = result_point[ALL_CLASSES_NAME].f_measure
            result.all_classes_f_measure_curve.append(all_classes_f_measure)

            for class_name in classes:
                result.f_measure_curve[class_name].append(result_point[class_name].f_measure)
                result.precision_curve[class_name].append(result_point[class_name].precision)
                result.recall_curve[class_name].append(result_point[class_name].recall)

            if all_classes_f_measure >= result.best_f_measure:
                result.best_f_measure = all_classes_f_measure
                result.best_threshold = nms_threshold
        return result

    def evaluate_classes(
        self, classes: List[str], iou_threshold: float, confidence_threshold: float
    ) -> Dict[str, _Metrics]:
        """Returns Dict of f_measure, precision and recall for each class.

        Args:
            classes (List[str]): List of classes to be evaluated.
            iou_threshold (float): IoU threshold to use for false negatives.
            confidence_threshold (float): Confidence threshold to use for false negatives.

        Returns:
            Dict[str, _Metrics]: The metrics (e.g. F-measure) for each class.
        """
        result: Dict[str, _Metrics] = {}

        all_classes_counters = _ResultCounters(0, 0, 0)

        if ALL_CLASSES_NAME in classes:
            classes.remove(ALL_CLASSES_NAME)
        for class_name in classes:
            metrics, counters = self.get_f_measure_for_class(
                class_name=class_name,
                iou_threshold=iou_threshold,
                confidence_threshold=confidence_threshold,
            )
            result[class_name] = metrics
            all_classes_counters.n_false_negatives += counters.n_false_negatives
            all_classes_counters.n_true += counters.n_true
            all_classes_counters.n_predicted += counters.n_predicted

        # for all classes
        result[ALL_CLASSES_NAME] = all_classes_counters.calculate_f_measure()
        return result

    def get_f_measure_for_class(
        self, class_name: str, iou_threshold: float, confidence_threshold: float
    ) -> Tuple[_Metrics, _ResultCounters]:
        """Get f_measure for specific class, iou threshold, and confidence threshold.

        In order to reduce the number of redundant iterations and allow for cleaner, more general code later on,
        all boxes are filtered at this stage by class and predicted boxes are filtered by confidence threshold

        Args:

            class_name (str): Name of the class for which the F measure is computed
            iou_threshold (float): IoU threshold
            confidence_threshold (float): Confidence threshold

        Returns:
            Tuple[_Metrics, _ResultCounters]: a structure containing the statistics (e.g. f_measure) and a structure
            containing the intermediated counters used to derive the stats (e.g. num. false positives)
        """
        class_ground_truth_boxes_per_image = self.__filter_class(self.ground_truth_boxes_per_image, class_name)
        confidence_predicted_boxes_per_image = self.__filter_confidence(
            self.prediction_boxes_per_image, confidence_threshold
        )
        class_predicted_boxes_per_image = self.__filter_class(confidence_predicted_boxes_per_image, class_name)
        if len(class_ground_truth_boxes_per_image) > 0:
            boxes_pair_per_class = _FMeasureCalculator(
                ground_truth_boxes_per_image=class_ground_truth_boxes_per_image,
                prediction_boxes_per_image=class_predicted_boxes_per_image,
            )
            result_counters = boxes_pair_per_class.get_counters(iou_threshold=iou_threshold)
            result_metrics = result_counters.calculate_f_measure()
            results = (result_metrics, result_counters)
        else:
            logger.warning("No ground truth images supplied for f-measure calculation.")
            # [f_measure, precision, recall, n_false_negatives, n_true, n_predicted]
            results = (_Metrics(0.0, 0.0, 0.0), _ResultCounters(0, 0, 0))
        return results

    @staticmethod
    def __get_critical_nms(
        boxes_per_image: List[List[Tuple[float, float, float, float, str, float]]], cross_class_nms: bool = False
    ) -> List[List[float]]:
        """Return list of critical NMS values for each box in each image.

        Maps each predicted box to the highest nms-threshold which would suppress that box, aka the smallest
        nms_threshold before the box disappears.
        Having these values allows us to later filter by nms-threshold in O(n) rather than O(n**2)
        Highest losing iou, holds the value of the highest iou that a box has with any
        other box of the same class and higher confidence score.

        Args:
            boxes_per_image (List[List[Tuple[float, float, float, float, str, float]]]): List of predicted boxes per
                image.
                a box: [x1: float, y1, x2, y2, class: str, score: float]
                boxes_per_image: [box1, box2, …]
            cross_class_nms (bool): Whether to use cross class NMS.

        Returns:
            List[List[float]]: List of critical NMS values for each box in each image.
        """
        critical_nms_per_image = []
        for boxes in boxes_per_image:
            critical_nms_per_box = []
            for box1 in boxes:
                highest_losing_iou = 0.0
                for box2 in boxes:
                    iou = bounding_box_intersection_over_union(box1, box2)
                    if (
                        (cross_class_nms or box1[FMeasure.box_class_index] == box2[FMeasure.box_class_index])
                        # TODO boxes Tuple should be refactored to dataclass.
                        and box1[FMeasure.box_score_index] < box2[FMeasure.box_score_index]  # type: ignore[operator]
                        and iou > highest_losing_iou
                    ):
                        highest_losing_iou = iou
                critical_nms_per_box.append(highest_losing_iou)
            critical_nms_per_image.append(critical_nms_per_box)
        return critical_nms_per_image

    @staticmethod
    def __filter_nms(
        boxes_per_image: List[List[Tuple[float, float, float, float, str, float]]],
        critical_nms: List[List[float]],
        nms_threshold: float,
    ) -> List[List[Tuple[float, float, float, float, str, float]]]:
        """Filters out predicted boxes whose critical nms is higher than the given nms_threshold.

        Args:
            boxes_per_image (List[List[Tuple[float, float, float, float, str, float]]]): List of boxes per image.
                a box: [x1: float, y1, x2, y2, class: str, score: float]
                boxes_per_image: [box1, box2, …]
            critical_nms (List[List[float]]): List of list of critical nms for each box in each image
            nms_threshold (float): NMS threshold used for filtering

        Returns:
            List[List[Tuple[float, float, float, float, str, float]]]: List of list of filtered boxes in each image
        """
        new_boxes_per_image = []
        for boxes, boxes_nms in zip(boxes_per_image, critical_nms):
            new_boxes = []
            for box, nms in zip(boxes, boxes_nms):
                if nms < nms_threshold:
                    new_boxes.append(box)
            new_boxes_per_image.append(new_boxes)
        return new_boxes_per_image

    @staticmethod
    def __filter_class(
        boxes_per_image: List[List[Tuple[float, float, float, float, str, float]]], class_name: str
    ) -> List[List[Tuple[float, float, float, float, str, float]]]:
        """Filters boxes to only keep members of one class.

        Args:
            boxes_per_image (List[List[Tuple[float, float, float, float, str, float]]]): a list of lists of boxes
            class_name (str): Name of the class for which the boxes are filtered

        Returns:
            List[List[Tuple[float, float, float, float, str, float]]]: a list of lists of boxes
        """
        filtered_boxes_per_image = []
        for boxes in boxes_per_image:
            filtered_boxes = []
            for box in boxes:
                # TODO boxes Tuple should be refactored to dataclass. This way we can access box.class
                if box[FMeasure.box_class_index].lower() == class_name.lower():  # type: ignore[union-attr]
                    filtered_boxes.append(box)
            filtered_boxes_per_image.append(filtered_boxes)
        return filtered_boxes_per_image

    @staticmethod
    def __filter_confidence(
        boxes_per_image: List[List[Tuple[float, float, float, float, str, float]]], confidence_threshold: float
    ) -> List[List[Tuple[float, float, float, float, str, float]]]:
        """Filters boxes to only keep ones with higher confidence than a given confidence threshold.

        Args:
            boxes_per_image (List[List[Tuple[float, float, float, float, str, float]]]):
                a box: [x1: float, y1, x2, y2, class: str, score: float]
                boxes_per_image: [box1, box2, …]
            confidence_threshold (float): Confidence threshold

        Returns:
            List[List[Tuple[float, float, float, float, str, float]]]: Boxes with higher confidence than the given
                threshold.
        """
        filtered_boxes_per_image = []
        for boxes in boxes_per_image:
            filtered_boxes = []
            for box in boxes:
                if float(box[FMeasure.box_score_index]) > confidence_threshold:
                    filtered_boxes.append(box)
            filtered_boxes_per_image.append(filtered_boxes)
        return filtered_boxes_per_image

    def get_counters(self, iou_threshold: float) -> _ResultCounters:
        """Return counts of true positives, false positives and false negatives for a given iou threshold.

        For each image (the loop), compute the number of false negatives, the number of predicted boxes, and the number
        of ground truth boxes, then add each value to its corresponding counter

        Args:
            iou_threshold (float): IoU threshold

        Returns:
            _ResultCounters: Structure containing the number of false negatives, true positives and predictions.
        """
        n_false_negatives = 0
        n_true = 0
        n_predicted = 0
        for ground_truth_boxes, predicted_boxes in zip(
            self.ground_truth_boxes_per_image, self.prediction_boxes_per_image
        ):
            n_true += len(ground_truth_boxes)
            n_predicted += len(predicted_boxes)
            if len(predicted_boxes) > 0:
                if len(ground_truth_boxes) > 0:
                    iou_matrix = get_iou_matrix(ground_truth_boxes, predicted_boxes)
                    n_false_negatives += get_n_false_negatives(iou_matrix, iou_threshold)
            else:
                n_false_negatives += len(ground_truth_boxes)
        return _ResultCounters(n_false_negatives, n_true, n_predicted)


class FMeasure(IPerformanceProvider):
    """Computes the f-measure (also known as F1-score) for a resultset.

    The f-measure is typically used in detection (localization) tasks to obtain a single number that balances precision
    and recall.

    To determine whether a predicted box matches a ground truth box an overlap measured
    is used based on a minimum
    intersection-over-union (IoU), by default a value of 0.5 is used.

    In addition spurious results are eliminated by applying non-max suppression (NMS) so that two predicted boxes with
    IoU > threshold are reduced to one. This threshold can be determined automatically by setting `vary_nms_threshold`
    to True.

    Args:
        resultset (ResultSetEntity) :ResultSet entity used for calculating the F-Measure
        vary_confidence_threshold (bool): if True the maximal F-measure is determined by optimizing for different
            confidence threshold values Defaults to False.
        vary_nms_threshold (bool): if True the maximal F-measure is determined by optimizing for different NMS threshold
            values. Defaults to False.
        cross_class_nms (bool): Whether non-max suppression should be applied cross-class. If True this will eliminate
            boxes with sufficient overlap even if they are from different classes. Defaults to False.

    Raises:
        ValueError: if prediction dataset and ground truth dataset are empty
    """

    def __init__(
        self,
        resultset: ResultSetEntity,
        vary_confidence_threshold: bool = False,
        vary_nms_threshold: bool = False,
        cross_class_nms: bool = False,
    ):

        ground_truth_dataset: DatasetEntity = resultset.ground_truth_dataset
        prediction_dataset: DatasetEntity = resultset.prediction_dataset

        if len(prediction_dataset) == 0 or len(ground_truth_dataset) == 0:
            raise ValueError("Cannot compute the F-measure of an empty result set.")

        ground_truth_dataset.sort_items()
        prediction_dataset.sort_items()

        labels = resultset.model.configuration.get_label_schema().get_labels(include_empty=False)
        classes = [label.name for label in labels]
        boxes_pair = _FMeasureCalculator(
            FMeasure.__get_boxes_from_dataset_as_list(ground_truth_dataset, labels),
            FMeasure.__get_boxes_from_dataset_as_list(prediction_dataset, labels),
        )
        result = boxes_pair.evaluate_detections(
            result_based_nms_threshold=vary_nms_threshold,
            classes=classes,
            cross_class_nms=cross_class_nms,
        )
        self._f_measure = ScoreMetric(name="f-measure", value=result.best_f_measure)
        self._f_measure_per_label: Dict[LabelEntity, ScoreMetric] = {}
        for label in labels:
            self.f_measure_per_label[label] = ScoreMetric(
                name=label.name, value=result.best_f_measure_per_class[label.name]
            )

        self._f_measure_per_confidence: Optional[CurveMetric] = None
        self._best_confidence_threshold: Optional[ScoreMetric] = None

        if vary_confidence_threshold:
            self._f_measure_per_confidence = CurveMetric(
                name="f-measure per confidence",
                xs=list(np.arange(*boxes_pair.confidence_range)),
                ys=result.per_confidence.all_classes_f_measure_curve,
            )
            self._best_confidence_threshold = ScoreMetric(
                name="Optimal confidence threshold",
                value=result.per_confidence.best_threshold,
            )

        self._f_measure_per_nms: Optional[CurveMetric] = None
        self._best_nms_threshold: Optional[ScoreMetric] = None

        if vary_nms_threshold and result.per_nms is not None:
            self._f_measure_per_nms = CurveMetric(
                name="f-measure per nms",
                xs=list(np.arange(*boxes_pair.nms_range)),
                ys=result.per_nms.all_classes_f_measure_curve,
            )
            self._best_nms_threshold = ScoreMetric(name="Optimal nms threshold", value=result.per_nms.best_threshold)

    box_class_index = 4
    box_score_index = 5

    @property
    def f_measure(self) -> ScoreMetric:
        """Returns the f-measure as ScoreMetric."""
        return self._f_measure

    @property
    def f_measure_per_label(self) -> Dict[LabelEntity, ScoreMetric]:
        """Returns the f-measure per label as dictionary (Label -> ScoreMetric)."""
        return self._f_measure_per_label

    @property
    def f_measure_per_confidence(self) -> Optional[CurveMetric]:
        """Returns the curve for f-measure per confidence as CurveMetric if exists."""
        return self._f_measure_per_confidence

    @property
    def best_confidence_threshold(self) -> Optional[ScoreMetric]:
        """Returns best confidence threshold as ScoreMetric if exists."""
        return self._best_confidence_threshold

    @property
    def f_measure_per_nms(self) -> Optional[CurveMetric]:
        """Returns the curve for f-measure per nms threshold as CurveMetric if exists."""
        return self._f_measure_per_nms

    @property
    def best_nms_threshold(self) -> Optional[ScoreMetric]:
        """Returns the best NMS threshold as ScoreMetric if exists."""
        return self._best_nms_threshold

    def get_performance(self) -> Performance:
        """Returns the performance which consists of the F-Measure score and the dashboard metrics.

        Returns:
            Performance: Performance object containing the F-Measure score and the dashboard metrics.
        """
        score = self.f_measure
        dashboard_metrics: List[MetricsGroup] = []
        dashboard_metrics.append(
            BarMetricsGroup(
                metrics=list(self.f_measure_per_label.values()),
                visualization_info=BarChartInfo(
                    name="F-measure per label",
                    palette=ColorPalette.LABEL,
                    visualization_type=VisualizationType.RADIAL_BAR,
                ),
            )
        )
        if self.f_measure_per_confidence is not None:
            dashboard_metrics.append(
                LineMetricsGroup(
                    metrics=[self.f_measure_per_confidence],
                    visualization_info=LineChartInfo(
                        name="F-measure per confidence",
                        x_axis_label="Confidence threshold",
                        y_axis_label="F-measure",
                    ),
                )
            )

        if self.best_confidence_threshold is not None:
            dashboard_metrics.append(
                TextMetricsGroup(
                    metrics=[self.best_confidence_threshold],
                    visualization_info=TextChartInfo(
                        name="Optimal confidence threshold",
                    ),
                )
            )

        if self.f_measure_per_nms is not None:
            dashboard_metrics.append(
                LineMetricsGroup(
                    metrics=[self.f_measure_per_nms],
                    visualization_info=LineChartInfo(
                        name="F-measure per nms",
                        x_axis_label="NMS threshold",
                        y_axis_label="F-measure",
                    ),
                )
            )

        if self.best_nms_threshold is not None:
            dashboard_metrics.append(
                TextMetricsGroup(
                    metrics=[self.best_nms_threshold],
                    visualization_info=TextChartInfo(
                        name="Optimal nms threshold",
                    ),
                )
            )
        return Performance(score=score, dashboard_metrics=dashboard_metrics)

    @staticmethod
    def __get_boxes_from_dataset_as_list(
        dataset: DatasetEntity, labels: List[LabelEntity]
    ) -> List[List[Tuple[float, float, float, float, str, float]]]:
        """Return list of boxes from dataset.

        Explanation of output shape:
            a box: [x1: float, y1, x2, y2, class: str, score: float]
            boxes_per_image: [box1, box2, …]
            ground_truth_boxes_per_image: [boxes_per_image_1, boxes_per_image_2, boxes_per_image_3, …]

        Args:
            dataset (DatasetEntity): Dataset to get boxes from.
            labels (List[LabelEntity]): Labels to get boxes for.

        Returns:
            List[List[Tuple[float, float, float, float, str, float]]]: List of boxes for each image in the dataset.
        """
        boxes_per_image = []
        converted_types_to_box = set()
        label_names = {label.name for label in labels}
        for item in dataset:
            boxes: List[Tuple[float, float, float, float, str, float]] = []
            roi_as_box = Annotation(ShapeFactory.shape_as_rectangle(item.roi.shape), labels=[])
            for annotation in item.annotation_scene.annotations:
                shape_as_box = ShapeFactory.shape_as_rectangle(annotation.shape)
                box = shape_as_box.normalize_wrt_roi_shape(roi_as_box.shape)
                n_boxes_before = len(boxes)
                boxes.extend(
                    [
                        (box.x1, box.y1, box.x2, box.y2, label.name, label.probability)
                        for label in annotation.get_labels()
                        if label.name in label_names
                    ]
                )
                if not isinstance(annotation.shape, Rectangle) and len(boxes) > n_boxes_before:
                    converted_types_to_box.add(annotation.shape.__class__.__name__)
            boxes_per_image.append(boxes)
        if len(converted_types_to_box) > 0:
            logger.warning(
                f"The shapes of types {tuple(converted_types_to_box)} have been converted to their "
                f"full enclosing Box representation in order to compute the f-measure"
            )

        return boxes_per_image
